"""Collect empirical dispersion parameters."""

import collections
import copy
from typing import Dict, List, Optional, Union

from ..exceptions import InputError

## ==> Dispersion Aliases and Parameters <== ##

# The dashcoeff dict below defines the -D parameters for most of the DFT methods. Some 'd2' are
#   taken from already defined functionals in psi4. Other parameters taken from the
#   references indicated in the "citation" parameter. The remainder of the parameters are
#   from http://toc.uni-muenster.de/DFTD3/ on September 25, 2012, with the dict keys
#   translated from Turbomole to Psi4 functional names.
#   Note: most DSD-functionals have their dispersion correction defined in dict_dh_funcs.py
dashcoeff = {
    "d2": {
        "formal": "D2",
        "alias": ["d"],
        "description": "    Grimme's -D2 Dispersion Correction",
        "citation": "    Grimme, S. (2006), J. Comp. Chem., 27: 1787-1799\n",
        "bibtex": "Grimme:2006:1787",
        "default": collections.OrderedDict([("s6", 1.0), ("alpha6", 20.0), ("sr6", 1.1)]),
        "definitions": {
            "blyp": {"params": {"s6": 1.2, "alpha6": 20.0, "sr6": 1.1}},
            "bp86": {"params": {"s6": 1.05, "alpha6": 20.0, "sr6": 1.1}},
            "b97": {"params": {"s6": 1.25, "alpha6": 20.0, "sr6": 1.1}},  # formerly b97-d
            "revpbe": {
                "params": {"s6": 1.25, "alpha6": 20.0, "sr6": 1.1},
                "citation": "    S. Grimme, J. Antony, S. Ehrlich, H. Krieg, J. Chem. Phys 132, 154104, 2010\n",
            },
            "pbe": {"params": {"s6": 0.75, "alpha6": 20.0, "sr6": 1.1}},
            "tpss": {"params": {"s6": 1.0, "alpha6": 20.0, "sr6": 1.1}},
            "b3lyp": {"params": {"s6": 1.05, "alpha6": 20.0, "sr6": 1.1}},
            "pbe0": {"params": {"s6": 0.6, "alpha6": 20.0, "sr6": 1.1}},
            "pw6b95": {"params": {"s6": 0.5, "alpha6": 20.0, "sr6": 1.1}},
            "tpss0": {
                "params": {"s6": 0.85, "alpha6": 20.0, "sr6": 1.1},
                "citation": "    S. Grimme, J. Antony, S. Ehrlich, H. Krieg, J. Chem. Phys 132, 154104, 2010\n",
            },
            "b2plyp": {"params": {"s6": 0.55, "alpha6": 20.0, "sr6": 1.1}},
            "b2gpplyp": {"params": {"s6": 0.4, "alpha6": 20.0, "sr6": 1.1}},
            "dsd-blyp": {"params": {"s6": 0.41, "alpha6": 60.0, "sr6": 1.1}},
            "core-dsd-blyp": {"params": {"s6": 0.41, "alpha6": 60.0, "sr6": 1.1}},
        },
    },
    "d3zeroatm": {
        "formal": "D3ATM",
        "alias": [],
        "description": "    Grimme's -D3 (zero-damping) Dispersion Correction with ATM",
        "citation": "    Grimme S.; Antony J.; Ehrlich S.; Krieg H. (2010), J. Chem. Phys., 132: 154104\n",
        "bibtex": "Grimme:2010:154104",
        "default": collections.OrderedDict(
            [("s6", 1.0), ("s8", 0.0), ("sr6", 1.0), ("alpha6", 14.0), ("sr8", 1.0), ("s9", 1.0)]
        ),
        "definitions": {
            # D3 parameters loaded from d3zero2b and from authoritative source below and s9 parameter added
        },
    },
    "d3zero2b": {
        "formal": "D3",
        "alias": ["d3", "d3zero", "d32b"],
        "description": "    Grimme's -D3 (zero-damping) Dispersion Correction",
        "citation": "    Grimme S.; Antony J.; Ehrlich S.; Krieg H. (2010), J. Chem. Phys., 132: 154104\n",
        "bibtex": "Grimme:2010:154104",
        "default": collections.OrderedDict([("s6", 1.0), ("s8", 0.0), ("sr6", 1.0), ("alpha6", 14.0), ("sr8", 1.0)]),
        "definitions": {
            # S. Grimme, J. Antony, S. Ehrlich, H. Krieg, J. Chem. Phys 132, 154104, 2010
            # 'b2plyp'      : {'s6': 0.5,  's8': 1.000, 'sr6': 1.332, 'alpha6': 14.0}, # superseded by a value below.
            "pw6b95": {"params": {"s6": 1.0, "s8": 0.862, "sr6": 1.532, "alpha6": 14.0, "sr8": 1.000}},
            "b97": {"params": {"s6": 1.0, "s8": 0.909, "sr6": 0.892, "alpha6": 14.0, "sr8": 1.000}},  # formerly b97-d
            "b3lyp": {"params": {"s6": 1.0, "s8": 1.703, "sr6": 1.261, "alpha6": 14.0, "sr8": 1.000}},
            "blyp": {"params": {"s6": 1.0, "s8": 1.682, "sr6": 1.094, "alpha6": 14.0, "sr8": 1.000}},
            "tpss0": {"params": {"s6": 1.0, "s8": 1.242, "sr6": 1.252, "alpha6": 14.0, "sr8": 1.000}},
            "pbe0": {"params": {"s6": 1.0, "s8": 0.928, "sr6": 1.287, "alpha6": 14.0, "sr8": 1.000}},
            "tpss": {"params": {"s6": 1.0, "s8": 1.105, "sr6": 1.166, "alpha6": 14.0, "sr8": 1.000}},
            "pbe": {"params": {"s6": 1.0, "s8": 0.722, "sr6": 1.217, "alpha6": 14.0, "sr8": 1.000}},
            "bp86": {"params": {"s6": 1.0, "s8": 1.683, "sr6": 1.139, "alpha6": 14.0, "sr8": 1.000}},
            # Later references
            "rpw86pbe": {
                "params": {"s6": 1.0, "s8": 0.901, "sr6": 1.224, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    S. Grimme, S. Ehrlich, L. Goerigk, J. Comput. Chem. 32, 1456-1465, 2011\n",
            },
            "b2plyp": {
                "params": {"s6": 0.64, "s8": 1.022, "sr6": 1.427, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, J. Chem. Theory Comput. 7, 291-309, 2011\n",
            },
            "b2gpplyp": {
                "params": {"s6": 0.56, "s8": 0.760, "sr6": 1.586, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, J. Chem. Theory Comput. 7, 291-309, 2011\n",
            },
            "pwpb95": {
                "params": {"s6": 0.82, "s8": 0.705, "sr6": 1.557, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, J. Chem. Theory Comput. 7, 291-309, 2011\n",
            },
            "dsd-blyp": {
                "params": {"s6": 0.50, "s8": 0.705, "sr6": 1.569, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, J. Chem. Theory Comput. 7, 291-309, 2011\n",
            },
            "bpbe": {
                "params": {"s6": 1.0, "s8": 2.033, "sr6": 1.087, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "opbe": {
                "params": {"s6": 1.0, "s8": 2.055, "sr6": 0.837, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "olyp": {
                "params": {"s6": 1.0, "s8": 1.764, "sr6": 0.806, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "mpwlyp": {
                "params": {"s6": 1.0, "s8": 1.098, "sr6": 1.239, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "bmk": {
                "params": {"s6": 1.0, "s8": 2.168, "sr6": 1.931, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "m05-2x": {
                "params": {"s6": 1.0, "s8": 0.00, "sr6": 1.417, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "m05": {
                "params": {"s6": 1.0, "s8": 0.595, "sr6": 1.373, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "m06-2x": {
                "params": {"s6": 1.0, "s8": 0.00, "sr6": 1.619, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "m06": {
                "params": {"s6": 1.0, "s8": 0.00, "sr6": 1.325, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "m06-l": {
                "params": {"s6": 1.0, "s8": 0.00, "sr6": 1.581, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "b3pw91": {
                "params": {"s6": 1.0, "s8": 1.775, "sr6": 1.176, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "bhlyp": {
                "params": {"s6": 1.0, "s8": 1.442, "sr6": 1.370, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "b1b95": {
                "params": {"s6": 1.0, "s8": 1.868, "sr6": 1.613, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "mpw1b95": {
                "params": {"s6": 1.0, "s8": 1.118, "sr6": 1.605, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "mpwb1k": {
                "params": {"s6": 1.0, "s8": 1.061, "sr6": 1.671, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "tpssh": {
                "params": {"s6": 1.0, "s8": 1.219, "sr6": 1.223, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "scan": {
                "params": {"s6": 1.0, "s8": 0.000, "sr6": 1.324, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    J.G. Brandenburg, J. E. Bates, J. Sun, J.P. Perdew, Phys. Rev. B 94, 115144, 2016\n",
            },
            "m11-l": {
                "params": {"s6": 1.0, "s8": 1.1129, "sr6": 2.3933, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "mn12-l": {
                "params": {"s6": 1.0, "s8": 0.9622, "sr6": 2.2329, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "n12": {
                "params": {"s6": 1.0, "s8": 2.3916, "sr6": 1.3493, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "sogga11-x": {
                "params": {"s6": 1.0, "s8": 1.8151, "sr6": 1.5431, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "m11": {
                "params": {"s6": 1.0, "s8": 0.6244, "sr6": 2.2300, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "mn12-sx": {
                "params": {"s6": 1.0, "s8": 0.8205, "sr6": 1.9572, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "n12-sx": {
                "params": {"s6": 1.0, "s8": 1.4713, "sr6": 1.4597, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "hse06": {
                "params": {"s6": 1.0, "s8": 0.1090, "sr6": 1.1290, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    J. Moellmann, S. Grimme, J. Chem. Phys. C 118, 7615-7621, 2014\n",
            },
            "wb97x": {
                "params": {"s6": 1.0, "s8": 1.0, "sr6": 1.281, "alpha6": 14.0, "sr8": 1.094},
                "citation": "    Y.-S. Lin, G.-D. Li, S.-P. Mao, J.-D. Chai, Chem. Theory Comput. 9, 263-272, 2013\n",
            },
            "pbehpbe": {
                "params": {"s6": 1.0, "s8": 1.4010, "sr6": 1.5703, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "rpbe": {
                "params": {"s6": 1.0, "s8": 0.5140, "sr6": 0.8720, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "xlyp": {
                "params": {"s6": 1.0, "s8": 0.7447, "sr6": 0.9384, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pw91p86": {
                "params": {"s6": 1.0, "s8": 0.8747, "sr6": 2.1040, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpwpw91": {
                "params": {"s6": 1.0, "s8": 1.9467, "sr6": 1.3725, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "hcth407": {
                "params": {"s6": 1.0, "s8": 2.7694, "sr6": 4.0426, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pkzb": {
                "params": {"s6": 1.0, "s8": 0.0000, "sr6": 0.6327, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "revtpss": {
                "params": {"s6": 1.0, "s8": 1.3666, "sr6": 1.3491, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "thcth": {
                "params": {"s6": 1.0, "s8": 0.5662, "sr6": 0.9320, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mn15-l": {
                "params": {"s6": 1.0, "s8": 0.0000, "sr6": 3.3388, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b3p86": {
                "params": {"s6": 1.0, "s8": 1.1961, "sr6": 1.1897, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b1p96": {
                "params": {"s6": 1.0, "s8": 1.1209, "sr6": 1.1815, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b1lyp": {
                "params": {"s6": 1.0, "s8": 1.9467, "sr6": 1.3725, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpw1lyp": {
                "params": {"s6": 1.0, "s8": 1.9529, "sr6": 2.0512, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpw1pw91": {
                "params": {"s6": 1.0, "s8": 1.4758, "sr6": 1.2892, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pw1pw": {
                "params": {"s6": 1.0, "s8": 1.1786, "sr6": 1.4968, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpw1kcis": {
                "params": {"s6": 1.0, "s8": 2.2917, "sr6": 1.7231, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpwkcis1k": {
                "params": {"s6": 1.0, "s8": 1.7553, "sr6": 1.4853, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pbeh1pbe": {
                "params": {"s6": 1.0, "s8": 1.0430, "sr6": 1.3719, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pbe1kcis": {
                "params": {"s6": 1.0, "s8": 1.7934, "sr6": 3.6355, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "x3lyp": {
                "params": {"s6": 1.0, "s8": 0.2990, "sr6": 1.0000, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "o3lyp": {
                "params": {"s6": 1.0, "s8": 1.8058, "sr6": 1.4060, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b97-1": {
                "params": {"s6": 1.0, "s8": 1.6418, "sr6": 3.7924, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b97-2": {
                "params": {"s6": 1.0, "s8": 2.4661, "sr6": 1.7066, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b98": {
                "params": {"s6": 1.0, "s8": 1.9078, "sr6": 2.6895, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "hiss": {
                "params": {"s6": 1.0, "s8": 0.7615, "sr6": 1.3338, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "hse03": {
                "params": {"s6": 1.0, "s8": 1.0156, "sr6": 1.3944, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "revtpssh": {
                "params": {"s6": 1.0, "s8": 1.2504, "sr6": 1.3224, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "revtpss0": {
                "params": {"s6": 1.0, "s8": 1.0649, "sr6": 1.2881, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "tpss1kcis": {
                "params": {"s6": 1.0, "s8": 2.0902, "sr6": 1.7729, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "thcthhyb": {
                "params": {"s6": 1.0, "s8": 1.6302, "sr6": 1.5001, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "m08-hx": {
                "params": {"s6": 1.0, "s8": 0.0000, "sr6": 1.6247, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "lcwhpbe": {
                "params": {"s6": 1.0, "s8": 1.2797, "sr6": 1.3846, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpw2plyp": {
                "params": {"s6": 0.66, "s8": 0.7529, "sr6": 1.5527, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pbe0-dh": {
                "params": {"s6": 0.840, "s8": 0.748, "sr6": 1.394, "alpha6": 14.0, "sr8": 1.000},
                "citation": "    D. Bousquet, E. Bremond, J. C. Sancho-Garcia, I. Ciofini, C. Adamo, Theor. Chem. Acc. 134, 1602, 2015\n",
            },
            # https://www.chemie.uni-bonn.de/pctc/mulliken-center/software/dft-d3/functionals
            "pwb6k": {"params": {"s6": 1.0, "s8": 0.550, "sr6": 1.660, "alpha6": 14.0, "sr8": 1.000}},
            "revpbe": {"params": {"s6": 1.0, "s8": 1.010, "sr6": 0.923, "alpha6": 14.0, "sr8": 1.000}},
            "bop": {"params": {"s6": 1.0, "s8": 1.975, "sr6": 0.929, "alpha6": 14.0, "sr8": 1.000}},
            "otpss": {"params": {"s6": 1.0, "s8": 1.494, "sr6": 1.128, "alpha6": 14.0, "sr8": 1.000}},
            "pbe38": {"params": {"s6": 1.0, "s8": 0.998, "sr6": 1.333, "alpha6": 14.0, "sr8": 1.000}},
            "pbesol": {"params": {"s6": 1.0, "s8": 0.612, "sr6": 1.345, "alpha6": 14.0, "sr8": 1.000}},
            "revssb": {"params": {"s6": 1.0, "s8": 0.560, "sr6": 1.221, "alpha6": 14.0, "sr8": 1.000}},
            "ssb": {"params": {"s6": 1.0, "s8": 0.663, "sr6": 1.215, "alpha6": 14.0, "sr8": 1.000}},
            "cam-b3lyp": {"params": {"s6": 1.0, "s8": 1.217, "sr6": 1.378, "alpha6": 14.0, "sr8": 1.000}},
            "wpbe": {"params": {"s6": 1.0, "s8": 1.279, "sr6": 1.355, "alpha6": 14.0, "sr8": 1.000}},  # formerly lcwpbe
            "m06-hf": {"params": {"s6": 1.0, "s8": 0.00, "sr6": 1.446, "alpha6": 14.0, "sr8": 1.000}},
            "hcth120": {"params": {"s6": 1.0, "s8": 1.206, "sr6": 1.221, "alpha6": 14.0, "sr8": 1.000}},
            "ptpss": {"params": {"s6": 0.75, "s8": 0.879, "sr6": 1.541, "alpha6": 14.0, "sr8": 1.000}},
            "revpbe0": {"params": {"s6": 1.0, "s8": 0.792, "sr6": 0.949, "alpha6": 14.0, "sr8": 1.000}},
            "revpbe38": {"params": {"s6": 1.0, "s8": 0.862, "sr6": 1.021, "alpha6": 14.0, "sr8": 1.000}},
            # unreferenced
            "hf": {"params": {"s6": 1.0, "s8": 1.746, "sr6": 1.158, "alpha6": 14.0, "sr8": 1.000}},
        },
    },
    "d3bjatm": {
        "formal": "D3(BJ)ATM",
        "alias": [],
        "description": "    Grimme's -D3 (BJ-damping) Dispersion Correction with ATM",
        "citation": "    Grimme S.; Ehrlich S.; Goerigk L. (2011), J. Comput. Chem., 32: 1456\n",
        "bibtex": "Grimme:2011:1456",
        "default": collections.OrderedDict([("s6", 1.0), ("s8", 1.0), ("a1", 0.0), ("a2", 1.0), ("s9", 1.0)]),
        "definitions": {
            # D3 parameters loaded from d3bj2b and from authoritative source below and s9 parameter added
        },
    },
    "d3bj2b": {
        "formal": "D3(BJ)",
        "alias": ["d3bj"],
        "description": "    Grimme's -D3 (BJ-damping) Dispersion Correction",
        "citation": "    Grimme S.; Ehrlich S.; Goerigk L. (2011), J. Comput. Chem., 32: 1456\n",
        "bibtex": "Grimme:2011:1456",
        "default": collections.OrderedDict([("s6", 1.0), ("s8", 1.0), ("a1", 0.0), ("a2", 1.0)]),
        "definitions": {
            # S. Grimme, S. Ehrlich, L. Goerigk, J. Comput. Chem. 7, 3297-3305, 2011
            "pbe": {"params": {"s6": 1.000, "s8": 0.7875, "a1": 0.4289, "a2": 4.4407}},
            "revpbe": {"params": {"s6": 1.000, "s8": 2.3550, "a1": 0.5238, "a2": 3.5016}},
            "blyp": {"params": {"s6": 1.000, "s8": 2.6996, "a1": 0.4298, "a2": 4.2359}},
            "bp86": {"params": {"s6": 1.000, "s8": 3.2822, "a1": 0.3946, "a2": 4.8516}},
            "rpw86pbe": {"params": {"s6": 1.000, "s8": 1.3845, "a1": 0.4613, "a2": 4.5062}},
            "b97": {"params": {"s6": 1.000, "s8": 2.2609, "a1": 0.5545, "a2": 3.2297}},  # formerly b97-d
            "tpss": {"params": {"s6": 1.000, "s8": 1.9435, "a1": 0.4535, "a2": 4.4752}},
            "b3lyp": {"params": {"s6": 1.000, "s8": 1.9889, "a1": 0.3981, "a2": 4.4211}},
            "pw6b95": {"params": {"s6": 1.000, "s8": 0.7257, "a1": 0.2076, "a2": 6.3750}},
            "pbe0": {"params": {"s6": 1.000, "s8": 1.2177, "a1": 0.4145, "a2": 4.8593}},
            "tpss0": {"params": {"s6": 1.000, "s8": 1.2576, "a1": 0.3768, "a2": 4.5865}},
            # 'b2plyp'      : {'params': {'s6': 0.500, 's8':  1.0860, 'a1':  0.3451, 'a2': 4.7735}}, # superseded by a value below.
            "hf": {"params": {"s6": 1.000, "s8": 0.9171, "a1": 0.3385, "a2": 2.8830}},
            # Later references
            "bpbe": {
                "params": {"s6": 1.000, "s8": 4.0728, "a1": 0.4567, "a2": 4.3908},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "opbe": {
                "params": {"s6": 1.000, "s8": 3.3816, "a1": 0.5512, "a2": 2.9444},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "olyp": {
                "params": {"s6": 1.000, "s8": 2.6205, "a1": 0.5299, "a2": 2.8065},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "mpwlyp": {
                "params": {"s6": 1.000, "s8": 2.0077, "a1": 0.4831, "a2": 4.5323},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "b3pw91": {
                "params": {"s6": 1.000, "s8": 2.8524, "a1": 0.4312, "a2": 4.4693},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "bhlyp": {
                "params": {"s6": 1.000, "s8": 1.0354, "a1": 0.2793, "a2": 4.9615},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "b1b95": {
                "params": {"s6": 1.000, "s8": 1.4507, "a1": 0.2092, "a2": 5.5545},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "mpw1b95": {
                "params": {"s6": 1.000, "s8": 1.0508, "a1": 0.1955, "a2": 6.4177},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "mpwb1k": {
                "params": {"s6": 1.000, "s8": 0.9499, "a1": 0.1474, "a2": 6.6223},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "tpssh": {
                "params": {"s6": 1.000, "s8": 2.2382, "a1": 0.4529, "a2": 4.6550},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "bmk": {
                "params": {"s6": 1.000, "s8": 2.0860, "a1": 0.1940, "a2": 5.9197},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "b2plyp": {
                "params": {"s6": 0.640, "s8": 0.9147, "a1": 0.3065, "a2": 5.0570},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "b2gpplyp": {
                "params": {"s6": 0.560, "s8": 0.2597, "a1": 0.0000, "a2": 6.3332},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "pwpb95": {
                "params": {"s6": 0.820, "s8": 0.2904, "a1": 0.0000, "a2": 7.3141},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "dsd-blyp": {
                "params": {"s6": 0.500, "s8": 0.2130, "a1": 0.0000, "a2": 6.0519},
                "citation": "    L. Goerigk, S. Grimme, Phys. Chem. Chem. Phys. 13, 6670-6688, 2011\n",
            },
            "hse06": {
                "params": {"s6": 1.000, "s8": 2.3100, "a1": 0.3830, "a2": 5.6850},
                "citation": "    J. Moellmann, S. Grimme, J. Chem. Phys. C 118, 7615-7621, 2014\n",
            },
            "pw91": {
                "params": {"s6": 1.000, "s8": 1.9598, "a1": 0.6319, "a2": 4.5718},
                "citation": "    J.R. Reimers et al., Proc. Natl. Acad. Sci. USA 112, E6101-E6110, 2015\n",
            },
            "m11-l": {
                "params": {"s6": 1.000, "s8": 0.4446, "a1": 0.0000, "a2": 7.2496},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "mn12-l": {
                "params": {"s6": 1.000, "s8": 2.2674, "a1": 0.0000, "a2": 9.1494},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "n12": {
                "params": {"s6": 1.000, "s8": 4.8491, "a1": 0.3842, "a2": 5.3545},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "sogga11-x": {
                "params": {"s6": 1.000, "s8": 1.1426, "a1": 0.1330, "a2": 5.7381},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "m11": {
                "params": {"s6": 1.000, "s8": 2.8112, "a1": 0.0000, "a2": 10.1389},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "mn12-sx": {
                "params": {"s6": 1.000, "s8": 1.1674, "a1": 0.0983, "a2": 8.0259},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "n12-sx": {
                "params": {"s6": 1.000, "s8": 2.4900, "a1": 0.3283, "a2": 5.7898},
                "citation": "    L. Goerigk, J. Phys. Chem. Lett. 6, 3891-3896, 2015\n",
            },
            "scan": {
                "params": {"s6": 1.000, "s8": 0.0000, "a1": 0.5380, "a2": 5.4200},
                "citation": "    J.G. Brandenburg, J. E. Bates, J. Sun, J.P. Perdew, Phys. Rev. B 94, 115144, 2016\n",
            },
            "pbehpbe": {
                "params": {"s6": 1.000, "s8": 1.1152, "a1": 0.0000, "a2": 4.4407},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "rpbe": {
                "params": {"s6": 1.000, "s8": 0.8318, "a1": 0.1820, "a2": 4.0094},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "xlyp": {
                "params": {"s6": 1.000, "s8": 1.5669, "a1": 0.0809, "a2": 5.3166},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpwpw91": {
                "params": {"s6": 1.000, "s8": 0.3168, "a1": 0.3168, "a2": 4.7732},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "hcth407": {
                "params": {"s6": 1.000, "s8": 0.6490, "a1": 0.0000, "a2": 4.8162},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "revtpss": {
                "params": {"s6": 1.000, "s8": 1.4023, "a1": 0.4426, "a2": 4.4723},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "thcth": {
                "params": {"s6": 1.000, "s8": 1.2626, "a1": 0.0000, "a2": 5.6162},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b3p86": {
                "params": {"s6": 1.000, "s8": 3.3211, "a1": 0.4601, "a2": 4.9294},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b1p86": {
                "params": {"s6": 1.000, "s8": 3.5681, "a1": 0.4724, "a2": 4.9858},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b1lyp": {
                "params": {"s6": 1.000, "s8": 2.1167, "a1": 0.1986, "a2": 5.3875},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpw1pw91": {
                "params": {"s6": 1.000, "s8": 1.8744, "a1": 0.3342, "a2": 4.9819},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpw1kcis": {
                "params": {"s6": 1.000, "s8": 1.0893, "a1": 0.0576, "a2": 5.5314},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpwkcis1k": {
                "params": {"s6": 1.000, "s8": 1.2875, "a1": 0.0855, "a2": 5.8961},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pbeh1pbe": {
                "params": {"s6": 1.000, "s8": 1.4877, "a1": 0.0000, "a2": 7.0385},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pbe1kcis": {
                "params": {"s6": 1.000, "s8": 0.7688, "a1": 0.0000, "a2": 6.2794},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "x3lyp": {
                "params": {"s6": 1.000, "s8": 1.5744, "a1": 0.2022, "a2": 5.4184},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "o3lyp": {
                "params": {"s6": 1.000, "s8": 1.8171, "a1": 0.0963, "a2": 5.9940},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b97-1": {
                "params": {"s6": 1.000, "s8": 0.4814, "a1": 0.0000, "a2": 6.2279},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b97-2": {
                "params": {"s6": 1.000, "s8": 0.9448, "a1": 0.0000, "a2": 5.4603},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "b98": {
                "params": {"s6": 1.000, "s8": 0.7086, "a1": 0.0000, "a2": 6.0672},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "hiss": {
                "params": {"s6": 1.000, "s8": 1.6112, "a1": 0.0000, "a2": 7.3539},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "hse03": {
                "params": {"s6": 1.000, "s8": 1.1243, "a1": 0.0000, "a2": 6.8889},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "revtpssh": {
                "params": {"s6": 1.000, "s8": 1.4076, "a1": 0.2660, "a2": 5.3761},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "revtpss0": {
                "params": {"s6": 1.000, "s8": 1.6151, "a1": 0.2218, "a2": 5.7985},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "tpss1kcis": {
                "params": {"s6": 1.000, "s8": 1.0542, "a1": 0.0000, "a2": 6.0201},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "thcthhyb": {
                "params": {"s6": 1.000, "s8": 0.9585, "a1": 0.0000, "a2": 6.2303},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mn15": {
                "params": {"s6": 1.000, "s8": 2.0971, "a1": 0.7862, "a2": 7.5923},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "lcwhpbe": {
                "params": {"s6": 1.000, "s8": 1.1908, "a1": 0.2746, "a2": 5.3157},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "mpw2plyp": {
                "params": {"s6": 1.000, "s8": 0.6223, "a1": 0.4105, "a2": 5.0136},
                "citation": "    L. Goerigk, A. Hansen, C. Bauer, S. Ehrlich, A. Najibi, S. Grimme, Phys. Chem. Chem. Phys. 19, 32184-32215, 2017\n",
            },
            "pbe0-dh": {
                "params": {"s6": 0.840, "s8": 0.095, "a1": 0.000, "a2": 6.102},
                "citation": "    D. Bousquet, E. Bremond, J. C. Sancho-Garcia, I. Ciofini, C. Adamo, Theor. Chem. Acc. 134, 1602, 2015\n",
            },
            # https://www.chemie.u{'params': ni-bonn.de/pctc/mulliken-center/software/dft-d3/functiona}lsbj
            "bop": {"params": {"s6": 1.000, "s8": 3.295, "a1": 0.4870, "a2": 3.5043}},
            "cam-b3lyp": {"params": {"s6": 1.000, "s8": 2.0674, "a1": 0.3708, "a2": 5.4743}},
            "wpbe": {"params": {"s6": 1.000, "s8": 1.8541, "a1": 0.3919, "a2": 5.0897}},  # formerly lcwpbe
            "otpss": {"params": {"s6": 1.000, "s8": 2.7495, "a1": 0.4634, "a2": 4.3153}},
            "pbe38": {"params": {"s6": 1.000, "s8": 1.4623, "a1": 0.3995, "a2": 5.1405}},
            "pbesol": {"params": {"s6": 1.000, "s8": 2.9491, "a1": 0.4466, "a2": 6.1742}},
            "ptpss": {"params": {"s6": 0.750, "s8": 0.2804, "a1": 0.000, "a2": 6.5745}},
            "pwb6k": {"params": {"s6": 1.000, "s8": 0.9383, "a1": 0.1805, "a2": 7.7627}},
            "revssb": {"params": {"s6": 1.000, "s8": 0.4389, "a1": 0.4720, "a2": 4.0986}},
            "ssb": {"params": {"s6": 1.000, "s8": -0.1744, "a1": -0.0952, "a2": 5.2170}},
            "hcth120": {"params": {"s6": 1.000, "s8": 1.0821, "a1": 0.3563, "a2": 4.3359}},
            "revpbe0": {"params": {"s6": 1.000, "s8": 1.7588, "a1": 0.4679, "a2": 3.7619}},
            "revpbe38": {"params": {"s6": 1.000, "s8": 1.4760, "a1": 0.4309, "a2": 3.9446}},
            # unreferenced
            # special HF/DFT with eBSSE correction
            "hf/mixed": {"params": {"s6": 1.000, "s8": 3.9027, "a1": 0.5607, "a2": 4.5622}},
            "hf/sv": {"params": {"s6": 1.000, "s8": 2.1849, "a1": 0.4249, "a2": 4.2783}},
            "hf/minis": {"params": {"s6": 1.000, "s8": 0.9841, "a1": 0.1702, "a2": 3.8506}},
            "b3lyp/6-31gd": {"params": {"s6": 1.000, "s8": 4.0672, "a1": 0.5014, "a2": 4.8409}},
            # special HF-D3-gCP-SRB/MINIX parametrization
            "hf3c": {"params": {"s6": 1.000, "s8": 0.8777, "a1": 0.4171, "a2": 2.9149}},
            # special HF-D3-gCP-SRB2/ECP-2G parametrization
            "hf3cv": {"params": {"s6": 1.000, "s8": 0.5022, "a1": 0.3063, "a2": 3.9856}},
            # special PBEh-D3-gCP/def2-mSVP parametrization
            "pbeh3c": {"params": {"s6": 1.000, "s8": 0.0000, "a1": 0.4860, "a2": 4.5000}},
            "core-dsd-blyp": {"params": {"s6": 0.500, "s8": 0.2130, "a1": 0.0000, "a2": 6.0519}},
        },
    },
    "d3mzeroatm": {
        "formal": "D3MATM",
        "alias": [],
        "description": "    Grimme's -D3 (zero-damping, short-range refitted) Dispersion Correction with ATM",
        "citation": "    Grimme S.; Antony J.; Ehrlich S.; Krieg H. (2010), J. Chem. Phys., 132: 154104\n"
        + "    Smith, D. G. A.; Burns, L. A.; Patkowski, K.; Sherrill, C. D. (2016), J. Phys. Chem. Lett.; 7: 2197\n",
        "bibtex": "Grimme:2010:154104",
        "default": collections.OrderedDict([("s6", 1.0), ("s8", 1.0), ("sr6", 1.0), ("beta", 1.0), ("s9", 1.0)]),
        "definitions": {
            # D3 parameters loaded from d3mzero2b and from authoritative source below and s9 parameter added
        },
    },
    "d3mzero2b": {
        "formal": "D3M",
        "alias": ["d3m", "d3mzero", "d3m2b"],
        "description": "    Grimme's -D3 (zero-damping, short-range refitted) Dispersion Correction",
        "citation": "    Grimme S.; Antony J.; Ehrlich S.; Krieg H. (2010), J. Chem. Phys., 132: 154104\n"
        + "    Smith, D. G. A.; Burns, L. A.; Patkowski, K.; Sherrill, C. D. (2016), J. Phys. Chem. Lett.; 7: 2197\n",
        "bibtex": "Grimme:2010:154104",
        "default": collections.OrderedDict([("s6", 1.0), ("s8", 1.0), ("sr6", 1.0), ("beta", 1.0)]),
        "definitions": {
            "b2plyp": {"params": {"s6": 0.640, "s8": 0.717543, "sr6": 1.313134, "beta": 0.016035}},
            "b3lyp": {"params": {"s6": 1.000, "s8": 1.532981, "sr6": 1.338153, "beta": 0.013988}},
            "b97": {"params": {"s6": 1.000, "s8": 1.020078, "sr6": 1.151808, "beta": 0.035964}},  # formerly b97-d
            "blyp": {"params": {"s6": 1.000, "s8": 1.841686, "sr6": 1.279637, "beta": 0.014370}},
            "bp86": {"params": {"s6": 1.000, "s8": 1.945174, "sr6": 1.233460, "beta": 0.000000}},
            "pbe": {"params": {"s6": 1.000, "s8": 0.000000, "sr6": 2.340218, "beta": 0.129434}},
            "pbe0": {"params": {"s6": 1.000, "s8": 0.000081, "sr6": 2.077949, "beta": 0.116755}},
            "wpbe": {"params": {"s6": 1.000, "s8": 1.280619, "sr6": 1.366361, "beta": 0.003160}},  # formerly lcwpbe
            "sapt0": {"params": {"s6": 1.000, "s8": 0.885517, "sr6": 1.383214, "beta": 0.075488}},  # JBS 01/2021
            "hf": {"params": {"s6": 1.000, "s8": 0.885517, "sr6": 1.383214, "beta": 0.075488}},  # JBS 01/2021
        },
    },
    "d3mbjatm": {
        "formal": "D3M(BJ)ATM",
        "alias": [],
        "description": "    Grimme's -D3 (BJ-damping, short-range refitted) Dispersion Correction with ATM",
        "citation": "    Grimme S.; Ehrlich S.; Goerigk L. (2011), J. Comput. Chem., 32: 1456\n"
        + "    Smith, D. G. A.; Burns, L. A.; Patkowski, K.; Sherrill, C. D. (2016), J. Phys. Chem. Lett.; 7: 2197\n",
        "bibtex": "Grimme:2011:1456",
        "default": collections.OrderedDict([("s6", 1.0), ("s8", 1.0), ("a1", 1.0), ("a2", 1.0), ("s9", 1.0)]),
        "definitions": {
            # D3 parameters loaded from d3mbj2b and from authoritative source below and s9 parameter added
        },
    },
    "d3mbj2b": {
        "formal": "D3M(BJ)",
        "alias": ["d3mbj"],
        "description": "    Grimme's -D3 (BJ-damping, short-range refitted) Dispersion Correction",
        "citation": "    Grimme S.; Ehrlich S.; Goerigk L. (2011), J. Comput. Chem., 32: 1456\n"
        + "    Smith, D. G. A.; Burns, L. A.; Patkowski, K.; Sherrill, C. D. (2016), J. Phys. Chem. Lett.; 7: 2197\n",
        "bibtex": "Grimme:2011:1456",
        "default": collections.OrderedDict([("s6", 1.0), ("s8", 1.0), ("a1", 1.0), ("a2", 1.0)]),
        "definitions": {
            "b2plyp": {"params": {"s6": 0.640, "s8": 0.672820, "a1": 0.486434, "a2": 3.656466}},
            "b3lyp": {"params": {"s6": 1.000, "s8": 1.466677, "a1": 0.278672, "a2": 4.606311}},
            "b97": {"params": {"s6": 1.000, "s8": 1.206988, "a1": 0.240184, "a2": 3.864426}},  # formerly b97-d
            "blyp": {"params": {"s6": 1.000, "s8": 1.875007, "a1": 0.448486, "a2": 3.610679}},
            "bp86": {"params": {"s6": 1.000, "s8": 3.140281, "a1": 0.821850, "a2": 2.728151}},
            "pbe": {"params": {"s6": 1.000, "s8": 0.358940, "a1": 0.012092, "a2": 5.938951}},
            "pbe0": {"params": {"s6": 1.000, "s8": 0.528823, "a1": 0.007912, "a2": 6.162326}},
            "wpbe": {"params": {"s6": 1.000, "s8": 0.906564, "a1": 0.563761, "a2": 3.593680}},  # formerly lcwpbe
            "sapt0": {"params": {"s6": 1.000, "s8": 0.713190, "a1": 0.079541, "a2": 3.627854}},  # JBS 01/2021
            "hf": {"params": {"s6": 1.000, "s8": 0.713190, "a1": 0.079541, "a2": 3.627854}},  # JBS 01/2021
        },
    },
    "d3opatm": {
        "formal": "D3(op)ATM",
        "alias": [],
        "description": "    D3 dispersion correction with optimized power damping function with ATM. Based on rational damping function and additional zero-damping like power function.",
        "citation": "    S. Grimme, S. Ehrlich, and L. Goerigk., Comput. Chem., 32:1456–1465, 2011. doi:10.1002/jcc.21759.\n"
        + "    Jonathon Witte, Narbe Mardirossian, Jeffrey B Neaton, and Martin Head-Gordon., J. Chem. Theory Comput., 13(5):2043–2052, 2017. doi:10.1021/acs.jctc.7b00176.\n",
        "bibtex": "Grimme:2011:1456",
        "default": collections.OrderedDict(
            [("s6", 1.0), ("s8", 1.0), ("a1", 1.0), ("a2", 1.0), ("beta", 0.0), ("s9", 1.0)]
        ),
        "definitions": {
            # D3 parameters loaded from d3op2b and from authoritative source below and s9 parameter added
        },
    },
    "d3op2b": {
        "formal": "D3(op)",
        "alias": ["d3op"],
        "description": "    D3 dispersion correction with optimized power damping function. Based on rational damping function and additional zero-damping like power function.",
        "citation": "    S. Grimme, S. Ehrlich, and L. Goerigk., Comput. Chem., 32:1456–1465, 2011. doi:10.1002/jcc.21759.\n"
        + "    Jonathon Witte, Narbe Mardirossian, Jeffrey B Neaton, and Martin Head-Gordon., J. Chem. Theory Comput., 13(5):2043–2052, 2017. doi:10.1021/acs.jctc.7b00176.\n",
        "bibtex": "Grimme:2011:1456",
        "default": collections.OrderedDict(
            [("s6", 1.0), ("s8", 1.0), ("a1", 1.0), ("a2", 1.0), ("beta", 0.0)],
        ),
        "definitions": {
            # will be loaded later
            # From https://github.com/dftd3/simple-dftd3/blob/main/assets/parameters.toml
        },
    },
    "nl": {
        "formal": "NL",
        "alias": [],
        "description": "    Grimme's -NL (DFT plus VV10 correlation) ",
        "citation": "    Hujo, W.; Grimme, S; (2011), J. Chem. Theory Comput.; 7:3866\n",
        "bibtex": "Hujo:2011:3866",
        "default": collections.OrderedDict([("b", 1.0), ("c", 0.0093)]),
        "definitions": {
            "blyp": {
                "params": {"b": 4.000, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme J. Chem. Theory Comput. 7, 3866-3871, 2011\n",
            },
            "hf": {
                "params": {"b": 3.900, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme J. Chem. Theory Comput. 7, 3866-3871, 2011\n",
            },  # not implemented
            "revpbe": {
                "params": {"b": 3.700, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme J. Chem. Theory Comput. 7, 3866-3871, 2011\n",
            },
            "revpbe38": {
                "params": {"b": 4.700, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme J. Chem. Theory Comput. 7, 3866-3871, 2011\n",
            },
            "b3lyp": {
                "params": {"b": 4.800, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme J. Chem. Theory Comput. 7, 3866-3871, 2011\n",
            },
            "b3pw91": {
                "params": {"b": 4.500, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme J. Chem. Theory Comput. 7, 3866-3871, 2011\n",
            },
            "revpbe0": {
                "params": {"b": 4.300, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme J. Chem. Theory Comput. 7, 3866-3871, 2011\n",
            },
            "bp86": {
                "params": {"b": 4.400, "c": 0.0093},
                "citation": "    M. K. Kesharwani, A. Karton, J.M. L. Martin, J. Chem. Theory Comput. 12, 444-454, 2016\n",
            },  # unclear if this is the real origin
            "pbe0": {
                "params": {"b": 6.900, "c": 0.0093},
                "citation": "    M. K. Kesharwani, A. Karton, J.M. L. Martin, J. Chem. Theory Comput. 12, 444-454, 2016\n",
            },  # unclear if this is the real origin
            "pbe": {
                "params": {"b": 6.400, "c": 0.0093},
                "citation": "    M. K. Kesharwani, A. Karton, J.M. L. Martin, J. Chem. Theory Comput. 12, 444-454, 2016\n",
            },  # unclear if this is the real origin
            "tpss0": {
                "params": {"b": 5.500, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme, J. Chem. Theory Comput. 9, 308-315, 2013\n",
            },
            "tpss": {
                "params": {"b": 5.000, "c": 0.0093},
                "citation": "    W. Hujo, S. Grimme, J. Chem. Theory Comput. 9, 308-315, 2013\n",
            },
            "b2gpplyp": {
                "params": {"b": 9.900, "c": 0.0093},
                "citation": "    M. K. Kesharwani, A. Karton, J.M. L. Martin, J. Chem. Theory Comput. 12, 444-454, 2016\n",
            },
            "b2plyp": {
                "params": {"b": 7.800, "c": 0.0093},
                "citation": "    J. Calbo, E. Orti, J. C. Sancho-Garcia, J. Arago, J. Chem. Theory Comput. 11, 932-939, 1015\n",
            },
            "pwpb95": {
                "params": {"b": 11.100, "c": 0.0093},
                "citation": "    F. Yu J. Chem. Theory Comput. 10, 4400-4407, 2014\n",
            },
            "revtpss": {
                "params": {"b": 5.465, "c": 0.0093},
                "citation": "    H. Kruse, P. Banas, J. Sponer, JCTC 2018, 10.1021/acs.jctc.8b00643 \n",
            },  # "just accepted"
        },
    },
    "d1": {
        "formal": "D1",
        "alias": [],
        "description": "    Grimme's -D1 Dispersion Correction",
        "citation": "    Grimme, S. (2004), J. Comp. Chem., 25: 1463-1473\n",
        "bibtex": "Grimme:2004:1463",
        "default": collections.OrderedDict([("s6", 1.0)]),
        "definitions": {"***": {"params": {"s6": 1.0}}},
    },
    "chg": {
        "formal": "CHG",
        "alias": [],
        "description": "    Chai and Head-Gordon Dispersion Correction",
        "citation": "    Chai, J.-D.; Head-Gordon, M. (2010), J. Chem. Phys., 132: 6615-6620\n",
        "bibtex": "Chai:2010:6615",
        "default": collections.OrderedDict([("s6", 1.0)]),
        "definitions": {"***": {"params": {"s6": 1.0}}},  # disp->d_ = 6.0;
    },
    "das2009": {
        "formal": "DAS2009",
        "alias": [],
        "description": "    Podeszwa and Szalewicz Dispersion Correction",
        "citation": "    Pernal, K.; Podeszwa, R.; Patkowski, K.; Szalewicz, K. (2009), Phys. Rev. Lett., 103: 263201\n",
        "bibtex": "Pernal:2009:263201",
        "default": collections.OrderedDict([("s6", 1.0)]),
        "definitions": {"***": {"params": {"s6": 1.0}}},
    },
    "das2010": {
        "formal": "DAS2010",
        "alias": [],
        "description": "    Podeszwa and Szalewicz Dispersion Correction",
        "citation": "    Podeszwa, R.; Pernal, K.; Patkowski, K.; Szalewicz, K. (2010), J. Phys. Chem. Lett., 1: 550\n",
        "bibtex": "Podeszwa:2010:550",
        "default": collections.OrderedDict([("s6", 1.0)]),
        "definitions": {"***": {"params": {"s6": 1.0}}},
    },
    "atmgr": {
        "formal": "ATM(GR)",
        "alias": [],
        "description": "    Grimme approximate Axilrod-Teller-Muto 3-body Dispersion Correction",
        "citation": "    Grimme S.; Antony J.; Ehrlich S.; Krieg H. (2010), J. Chem. Phys., 132: 154104\n",
        "bibtex": "Grimme:2010:154104",
        "default": collections.OrderedDict([("alpha6", 14.0)]),
        "definitions": {
            "***": {"params": {"alpha6": 14.0}, "citation": ""}  # alpha6 = 14 => damping parameter used is alp8 = 16
        },
    },
    "dmp2": {
        "formal": "DMP2",
        "alias": [],
        "description": "    Beran's MP2D Dispersion Correction for MP2",
        "citation": "    Rezac, J.; Greenwell, C.; Beran, G. (2018), J. Chem. Theory Comput., 14: 4711-4721\n",
        "bibtex": "Rezac:2018:4711",
        "doi": "10.1021/acs.jctc.8b00548",
        "default": collections.OrderedDict([("s8", 1.187), ("a1", 0.944), ("a2", 0.480), ("rcut", 0.72), ("w", 0.20)]),
        "definitions": {
            "mp2": {"params": {"s8": 1.187, "a1": 0.944, "a2": 0.480, "rcut": 0.72, "w": 0.20}}  # Rezac:2018:4711
        },
    },
    "d4bjeeqatm": {
        "formal": "D4(BJ,EEQ)ATM",
        "alias": ["d4", "d4bj", "d4(bj)"],
        "description": "    Grimme's -D4 (BJ-damping) Dispersion Correction with ATM",
        "citation": "    Caldeweyher, E.; Ehlert, S.; Hansen, A.; Neugebauer, H.; Spicher, S.; Bannwarth, C.; Grimmme, S., J. Chem. Phys. 150, 154122 (2019)\n",
        "bibtex": "Caldeweyher:2019:154122",
        "doi": "10.1063/1.5090222150",
        "default": collections.OrderedDict(
            [("a1", 1.0), ("a2", 1.0), ("alp", 16.0), ("s6", 1.0), ("s8", 1.0), ("s9", 1.0)]
        ),
        "definitions": {
            # D4 parameters loaded below from authoritative source below. Keep a couple for reference
            # "b3lyp": {"params": {"a1": 0.40868035, "a2": 4.53807137, "alp": 16.0, "s6": 1.0, "s8": 2.02929367, "s9": 1.0}},
            # "pbe":   {"params": {"a1": 0.38574991, "a2": 4.80688534, "alp": 16.0, "s6": 1.0, "s8": 0.95948085, "s9": 1.0}},
        },
    },
}


# for d3*atm, only skeleton entries with metadata defined above. below copies in parameters from d3*2b
for d in ["d3zero", "d3bj", "d3mzero", "d3mbj", "d3op"]:
    for k, v in dashcoeff[d + "2b"]["definitions"].items():
        dashcoeff[d + "atm"]["definitions"][k] = copy.deepcopy(dashcoeff[d + "2b"]["definitions"][k])
        dashcoeff[d + "atm"]["definitions"][k]["params"][
            "s9"
        ] = 1.0  # set twice: this one adds s9 to the params list for atm


def _get_d4bj_definitions() -> dict:
    """DFTD4 provides access to damping parameters on per functional basis.
    But we want all of them.

    We let DFTD4 take care of finding and loading the parameter file,
    but to get all parameters, we implement the logic to read those
    parameters ourselves again.

    This method is upstreamed in dftd4 >3.3.0 and could be replaced in the future.
    """

    try:
        from dftd4.parameters import get_data_file_name, load_data_base
    except ModuleNotFoundError:
        return {}

    def get_params(entry: dict, base: dict, defaults: list) -> dict:
        """Retrive the parameters from the data base, make sure the default
        values are applied correctly in the process. In case we have multiple
        defaults search for the first of the list defined for this method."""

        for default in defaults:
            try:
                params = base[default].copy()
                params.update(**entry[default])
                params.pop("mbd", None)
                params.pop("damping", None)
                return params
            except KeyError:
                continue

        raise KeyError("No entry for " + method + " in parameter data base")

    try:
        _data_base = load_data_base(get_data_file_name())
    except FileNotFoundError:
        return {}

    try:
        _defaults = _data_base["default"]["d4"]
        _base = _data_base["default"]["parameter"]["d4"]
        _parameters = _data_base["parameter"]
    except KeyError:
        return {}

    definitions = {}

    for method in _parameters:
        try:
            _entry = _parameters[method]["d4"]
            params = get_params(_entry, _base, _defaults)

            citation = params.pop("doi", None)
            # Make Psi4's citation style checker happy
            if citation is not None:
                citation = "    " + citation + "\n"
            definitions[method] = dict(
                params=params,
                citation=citation,
            )
        except KeyError:
            continue

    return definitions


dashcoeff["d4bjeeqatm"]["definitions"].update(_get_d4bj_definitions())

try:

    def _get_d3_definitions(dashlevel: str) -> dict:
        """
        Lucky for us dftd3 provides access to all damping parameters directly.

        However, we are not happy with the structure of the dict and
        therefore do some restructuring. The below solution tries to bend
        the data returned by dftd3 into the right shape to make qcng happy.
        """

        from dftd3.parameters import get_all_damping_params

        # The names here are the subset allowed by qcng with the names used in dftd3
        allowed = {
            "bj": ["a1", "a2", "s6", "s8", "s9"],
            "zero": ["rs6", "rs8", "alp", "s6", "s8", "s9"],
            "mbj": ["a1", "a2", "s6", "s8", "s9"],
            "mzero": ["rs6", "rs8", "alp", "s6", "s8", "bet", "s9"],
            "op": ["a1", "a2", "s6", "s8", "bet", "s9"],
        }
        # mapping from dftd3 to qcng names, also, we have to reverse it later again
        rename = {
            "rs6": "sr6",
            "rs8": "sr8",
            "alp": "alpha6",
            "bet": "beta",
        }

        # FIXME: dict comprehension at its finest
        #
        # 1. insert an additional level each method
        # 2. filter parameter names accepted by qcng
        # 3. map from dftd3 to qcng names
        # 4. return the whole mess
        return {
            key: {
                "params": {
                    rename.get(param, param): value for param, value in params.items() if param in allowed[dashlevel]
                }
            }
            for key, params in get_all_damping_params([dashlevel]).items()
        }

    dashcoeff["d3bjatm"]["definitions"].update(_get_d3_definitions("bj"))
    dashcoeff["d3zeroatm"]["definitions"].update(_get_d3_definitions("zero"))
    dashcoeff["d3mbjatm"]["definitions"].update(_get_d3_definitions("mbj"))
    dashcoeff["d3mzeroatm"]["definitions"].update(_get_d3_definitions("mzero"))
    dashcoeff["d3opatm"]["definitions"].update(_get_d3_definitions("op"))

    for d in ["d3zero", "d3bj", "d3mzero", "d3mbj", "d3op"]:
        for k, v in dashcoeff[d + "atm"]["definitions"].items():
            dashcoeff[d + "atm"]["definitions"][k]["params"][
                "s9"
            ] = 1.0  # set twice: this one establishes 1.0 if upstream is still 0.0 (pre v0.7.0 simple-dftd3)

except ModuleNotFoundError:
    pass


def get_dispersion_aliases():
    """Returns dict where keys are all (lowercased) strings that are
    interpretable as dashlevels, and values are the keys dealiased
    (d->d2) and deformalized (d3(bj)->d3bj) into valid dashcoeff keys.

    """
    alias = {}
    for dash, ddash in dashcoeff.items():
        alias[dash] = dash
        alias[ddash["formal"].lower()] = dash
        for al in ddash["alias"]:
            alias[al] = dash

    return alias


def from_arrays(
    name_hint: Optional[str] = None,
    level_hint: Optional[str] = None,
    param_tweaks: Union[List[float], Dict[str, float]] = None,
    dashcoeff_supplement: Optional[Dict[str, Dict]] = None,
    verbose: int = 1,
):
    """Use the three paths of empirical dispersion parameter information
    (DFT functional, dispersion correction level, and particular
    parameters) to populate the parameter array and validate a
    "functional-dispersion" label.

    Parameters
    ----------
    name_hint
        Name of functional (func only, func & disp, or disp only) for
        which to compute dispersion (e.g., blyp, BLYP-D2, blyp-d3bj,
        blyp-d3(bj), hf+d). Any or all parameters initialized from
        `dashcoeff[dashlevel][functional-without-dashlevel]` or
        `dashcoeff_supplement[dashlevel][functional-with-dashlevel]
        can be overwritten via `param_tweaks`.
    level_hint
        Name of dispersion correction to be applied (e.g., d, D2,
        d3(bj), das2010). Must be key in `dashcoeff` or "alias" or
        "formal" to one.
    param_tweaks
        Values for the same keys as `dashcoeff[dashlevel]['default']`
        (and same order if list) used to override any or all values
        initialized by `name_hint`.  Extra parameters will error.
    dashcoeff_supplement
        Dictionary of the same structure as `dashcoeff` that contains
        in "definitions" field full functional names, rather than
        fctl less dashlvl. Used to validate dict_builder fctls with
        dispersion or identify disp level from just `name_hint`.
    verbose
        Amount of printing.

    Returns
    -------
    dict
        Metadata defining dispersion calculation.

        dashlevel : {'d1', 'd2', 'd3zero2b', 'd3bj2b', 'd3mzero2b', 'd3mbj2b', 'd3op2b', 'd3zeroatm', 'd3bjatm', 'd3mzeroatm', 'd3mbjatm', 'd3opatm', 'chg', 'das2009', 'das2010', 'nl', "d4bjeeqatm"}
            Name (de-aliased, de-formalized, lowercase) of dispersion
            correction -- atom data, dispersion model, damping functional
            form -- to be applied. Resolved from `name_hint` and/or
            `level_hint` into a key of `dashparam.dashcoeff`.
        dashparams : dict
            Complete (number and parameter names vary by `dashlevel`)
            set of parameter values defining the flexible parts
            of `dashlevel`. Resolved into a complete set (keys of
            dashcoeff[dashlevel]['default']) from `name_hint` and/or
            `dashcoeff_supplement` and/or user `param_tweaks`.
        fctldash : str
            If `dashparams` for `dashlevel` corresponds to a defined,
            named, untweaked "functional-dashlevel" set, then that
            label. Otherwise, empty string.
        dashparams_citation : str
            If `dashparams` for `dashlevel` corresponds to a defined,
            named, untweaked "functional-dashlevel" set, and that
            parameter set has a literature citation, then that
            citation. Otherwise, empty string.

    Notes
    -----
    * No parameter is required, but somewhere the dispersion level and
      intended parameters must be clear. Explicit contradictory info will
      cause an error.
    * Function intended to be idempotent.

    """
    try:
        # try/except block retrofits Psi4 for pre-2b/atm-split in dashcoeff
        import psi4

        if "d3bj" in psi4.procrouting.empirical_dispersion._engine_can_do["dftd3"]:
            psi4.procrouting.empirical_dispersion._engine_can_do["dftd3"] = [
                "d2",
                "d3zero2b",
                "d3bj2b",
                "d3mzero2b",
                "d3mbj2b",
            ]
            for d in ["d3zero", "d3bj", "d3mzero", "d3mbj"]:
                psi4.procrouting.empirical_dispersion._capable_engines_for_disp[
                    d + "2b"
                ] = psi4.procrouting.empirical_dispersion._capable_engines_for_disp.pop(d)
    except ImportError:
        pass

    if verbose > 1:
        print(
            "empirical_dispersion_resources.from_arrays HINTS:",
            name_hint,
            level_hint,
            param_tweaks,
            bool(dashcoeff_supplement),
        )

    # << 0 >> prep
    if dashcoeff_supplement is not None:
        supplement_dashlevel_lookup = {}
        for disp, ddisp in dashcoeff_supplement.copy().items():
            for func, params in ddisp["definitions"].items():
                try:
                    dashcoeff[disp]
                except KeyError:
                    # try/except block accommodates dashcoeff_supplement from pre-2b/atm-split in dashcoeff
                    disp = get_dispersion_aliases()[disp]
                    dashcoeff_supplement[disp] = ddisp
                if params["params"].keys() != dashcoeff[disp]["default"].keys():
                    if verbose > 2:
                        print(
                            "Warning: trouble in dict_builder def:",
                            func,
                            params["params"].keys(),
                            "!=",
                            dashcoeff[disp]["default"].keys(),
                        )
                else:
                    supplement_dashlevel_lookup[func] = disp

    # << 1 >> use name_hint and/or level_hint to determine intended dispersion level
    if name_hint is None and level_hint is None:
        raise InputError(
            """Can't guess -D level without name_hint ({}) or level_hint ({})""".format(name_hint, level_hint)
        )

    if level_hint is None:
        dashlevel_candidate_1 = None
    else:
        level_hint = level_hint.lower()
        try:
            dashlevel_candidate_1 = get_dispersion_aliases()[level_hint]
        except KeyError:
            raise InputError(
                """Requested -D correction level ({}) not among ({})""".format(level_hint, dashcoeff.keys())
            )

    if name_hint is None:
        dashlevel_candidate_2 = None
        name_key = None
        disp_params = {}
    else:
        name_hint = name_hint.lower()
        trial_split = name_hint.rsplit("-", 1)

        if name_hint in get_dispersion_aliases():
            dashlevel_candidate_2 = get_dispersion_aliases()[name_hint]
            if list(dashcoeff[dashlevel_candidate_2]["definitions"]) == ["***"]:
                # case disp-only fctl-indep: chg, atmgr
                name_key = "***"
                disp_params = dashcoeff[dashlevel_candidate_2]["definitions"][name_key]["params"]
            else:
                # case disp-only: d3, d3zero, d3(bj)
                name_key = None
                disp_params = {}
        elif (dashcoeff_supplement is not None) and name_hint in supplement_dashlevel_lookup:
            # case fctldisp: wb97x-d, hf+d
            dashlevel_candidate_2 = supplement_dashlevel_lookup[name_hint]
            name_key = name_hint
            disp_params = dashcoeff_supplement[dashlevel_candidate_2]["definitions"][name_hint]["params"]
        elif (
            len(trial_split) == 2
            and trial_split[1] in get_dispersion_aliases()
            and trial_split[0] in dashcoeff[get_dispersion_aliases()[trial_split[1]]]["definitions"]
        ):
            # case fctldisp: b3lyp-d3, b3lyp-d3zero, b3lyp-d3(bj)
            dashlevel_candidate_2 = get_dispersion_aliases()[trial_split[1]]
            name_key = trial_split[0]
            disp_params = dashcoeff[dashlevel_candidate_2]["definitions"][trial_split[0]]["params"]
        elif (
            len(trial_split) == 2
            and trial_split[1] in get_dispersion_aliases()
            and list(dashcoeff[get_dispersion_aliases()[trial_split[1]]]["definitions"]) == ["***"]
        ):
            # case fctldisp: asdf-chg, pbe-atmgr
            dashlevel_candidate_2 = get_dispersion_aliases()[trial_split[1]]
            name_key = "***"
            disp_params = dashcoeff[dashlevel_candidate_2]["definitions"][name_key]["params"]
        elif (level_hint is not None) and name_hint in dashcoeff[dashlevel_candidate_1]["definitions"]:
            # case fctl: b3lyp
            dashlevel_candidate_2 = None
            name_key = name_hint
            disp_params = dashcoeff[dashlevel_candidate_1]["definitions"][name_hint]["params"]
        elif name_hint == "":
            dashlevel_candidate_2 = None
            name_key = None
            disp_params = {}
        # elif ((level_hint is not None) and list(dashcoeff[get_dispersion_aliases()[level_hint]]['definitions']) == ['***']):
        #    dashlevel_candidate_2 = get_dispersion_aliases()[level_hint]
        #    name_key = '***'
        #    disp_params = dashcoeff[dashlevel_candidate_2]['definitions'][name_key]['params']
        else:
            # dashlevel_candidate_2 = None
            raise InputError("""Can't guess -D correction level from ({})""".format(name_hint))

    disp_params = copy.deepcopy(disp_params)

    if dashlevel_candidate_1 is None and dashlevel_candidate_2 is None:
        raise InputError(
            f"""Can't guess -D correction level from name_hint ({name_hint}) and level_hint ({level_hint})"""
        )
    elif dashlevel_candidate_1 is not None and dashlevel_candidate_2 is not None:
        if dashlevel_candidate_1 != dashlevel_candidate_2:
            raise InputError(
                f"""Inconsistent -D correction level ({dashlevel_candidate_2} != {dashlevel_candidate_1}) from name_hint ({name_hint}) and level_hint ({level_hint})"""
            )
    dashleveleff = dashlevel_candidate_1 or dashlevel_candidate_2

    allowed_params = dashcoeff[dashleveleff]["default"].keys()

    # << 2 >> use name_hint and/or param_tweaks to determine intended dispersion parameters
    if name_hint is None and param_tweaks is None:
        raise InputError(
            """Can't guess -D parameters without name_hint ({}) or param_tweaks ({})""".format(name_hint, param_tweaks)
        )

    if isinstance(param_tweaks, list):
        param_tweaks = dict(zip(allowed_params, param_tweaks))

    if param_tweaks is None:
        param_tweaks = {}

    if not set(param_tweaks.keys()).issubset(allowed_params):
        raise InputError(
            "Requested keys ({}) not among allowed ({}) for dispersion level ({})".format(
                list(param_tweaks.keys()), list(allowed_params), dashleveleff
            )
        )

    disp_params.update(param_tweaks)

    if disp_params.keys() != allowed_params:
        raise InputError(
            "Requested keys ({}) insufficient ({}) for dispersion level ({})".format(
                list(param_tweaks.keys()), list(allowed_params), dashleveleff
            )
        )

    # << 3 >> use final dashlevel and disp_params to determine if a defined "fctl-disp" label exists
    # * plucks any citation for the parameters from definition source
    # * if/elif chooses right label when some fctls have identical param sets
    if (
        (name_hint is not None)
        and (dashcoeff_supplement is not None)
        and (name_key in dashcoeff_supplement[dashleveleff]["definitions"])
        and (disp_params == dashcoeff_supplement[dashleveleff]["definitions"][name_key]["params"])
    ):
        citeff = dashcoeff_supplement[dashleveleff]["definitions"][name_key].get("citation", "")
        if name_key == "***":
            fctldasheff = ""
        else:
            fctldasheff = name_key
    elif name_hint not in [None, ""] and (disp_params == dashcoeff[dashleveleff]["definitions"][name_key]["params"]):
        citeff = dashcoeff[dashleveleff]["definitions"][name_key].get("citation", "")
        if name_key == "***":
            fctldasheff = dashcoeff[dashleveleff]["formal"].lower()
        else:
            fctldasheff = "-".join([name_key, dashcoeff[dashleveleff]["formal"].lower()])
    else:
        if dashcoeff_supplement is not None:
            for func, params in dashcoeff_supplement[dashleveleff]["definitions"].items():
                if disp_params == params["params"]:
                    fctldasheff = func
                    citeff = params.get("citation", "")
                    break

        for func, params in dashcoeff[dashleveleff]["definitions"].items():
            if disp_params == params["params"]:
                fctldasheff = "-".join([func, dashcoeff[dashleveleff]["formal"].lower()])
                citeff = params.get("citation", "")
                break
        else:
            fctldasheff = ""
            citeff = ""

    # TODO right now fctldasheff is empty if undefined. use '-dash' or 'custom dash' instead?
    # TODO right now citation is empty if undefined. remove key or use None or False instead?

    if verbose > 1:
        print(
            f"empirical_dispersion_resources.from_arrays RESOLVED: dashlevel={dashleveleff}, dashparams={disp_params}, fctldash={fctldasheff}, dashparams_citation={citeff}"
        )

    return {
        "dashlevel": dashleveleff,
        "dashparams": disp_params,
        "fctldash": fctldasheff,
        "dashparams_citation": citeff,
    }



from typing import Any, Dict, Tuple

from qcengine.exceptions import InputError

# List of XC functionals known to NWChem
xc_functionals = [
    "acm",
    "b3lyp",
    "beckehandh",
    "pbe0",
    "becke97",
    "becke97-1",
    "becke97-2",
    "becke97-3",
    "becke97-d",
    "becke98",
    "hcth",
    "hcth120",
    "hcth147",
    "hcth407",
    "becke97gga1",
    "hcth407p",
    "mpw91",
    "mpw1k",
    "xft97",
    "cft97",
    "ft97",
    "op",
    "bop",
    "pbeop",
    "xpkzb99",
    "cpkzb99",
    "xtpss03",
    "ctpss03",
    "xctpssh",
    "b1b95",
    "bb1k",
    "mpw1b95",
    "mpwb1k",
    "pw6b95",
    "pwb6k",
    "m05",
    "m05-2x",
    "vs98",
    "m06",
    "m06-hf",
    "m06-L",
    "m06-2x",
    "hfexch",
    "becke88",
    "xperdew91",
    "xpbe96",
    "gill96",
    "lyp",
    "perdew81",
    "perdew86",
    "perdew91",
    "cpbe96",
    "pw91lda",
    "slater",
    "vwn_1",
    "vwn_2",
    "vwn_3",
    "vwn_4",
    "vwn_5",
    "vwn_1_rpa",
    "xtpss03",
    "ctpss03",
    "bc95",
    "xpw6b95",
    "xpwb6k",
    "xm05",
    "xm05-2x",
    "cpw6b95",
    "cpwb6k",
    "cm05",
    "cm05-2x",
    "xvs98",
    "cvs98",
    "xm06-L",
    "xm06-hf",
    "xm06",
    "xm06-2x",
    "cm06-L",
    "cm06-hf",
    "cm06",
    "cm06-2x",
]


def muster_modelchem(method: str, derint: int, use_tce: bool) -> Tuple[str, Dict[str, Any]]:
    """Converts the QC method into NWChem keywords

    Args:
        method (str): Name of the QC method to use
        derint (str): Index of the run type
        use_tce (bool): Whether to use the Tensor Contraction Engine
    Returns:
        (str): Task command for NWChem
        (dict): Any options for NWChem
    """

    # Standardize the method name
    method = method.lower()
    opts = {}

    # Map the run type to
    runtyp = {"energy": "energy", "gradient": "gradient", "hessian": "hessian", "properties": "property"}[derint]

    # Write out the theory directive
    if method == "nwchem":
        mdccmd = ""

    elif method in ["scf", "hf"]:
        mdccmd = f"task scf {runtyp}\n\n"

    elif method == "mp2":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__mp2"] = True
        else:
            mdccmd = f"task mp2 {runtyp}\n\n"

    elif method == "mp3":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__mp3"] = True

    elif method == "mp4":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__mp4"] = True

    elif method == "ccd":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__ccd"] = True

    elif method == "ccsd":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__ccsd"] = True
        else:
            mdccmd = f"task ccsd {runtyp}\n\n"

    elif method == "ccsd+t(ccsd)":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__ccsd(t)"] = True
        else:
            mdccmd = f"task ccsd+t(ccsd) {runtyp}\n\n"

    elif method == "ccsd(t)":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__ccsd(t)"] = True
        else:
            mdccmd = f"task ccsd(t) {runtyp}\n\n"

    elif method == "ccsdt":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__ccsdt"] = True
        else:
            mdccmd = f"task ccsdt {runtyp}\n\n"

    elif method == "tddft":
        mdccmd = f"task tddft {runtyp}\n\n"

    elif method in ["sodft", "direct_mp2", "rimp2", "mcscf", "selci", "md", "pspw", "band"]:
        raise InputError(f'Method "{method}" not yet supported by QCEngine')

    elif method == "tce":
        raise InputError(
            f"Do not specify TCE as a method. Instead specify the desired method " f'as a keyword and "qc_module=True".'
        )

    elif method.split()[0] in xc_functionals:
        opts["dft__xc"] = method
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__"] = "dft"
        else:
            mdccmd = f"task dft {runtyp}\n\n"

    elif method == "pbe":
        opts["dft__xc"] = "xpbe96 cpbe96"
        mdccmd = f"task dft {runtyp}\n\n"

    elif method == "b3lyp5":
        opts["dft__xc"] = "hfexch 0.2 slater 0.8 becke88 nonlocal 0.72 vwn_5 0.190 lyp 0.81"
        mdccmd = f"task dft {runtyp}\n\n"

    elif method == "dft":
        if use_tce:
            mdccmd = f"task tce {runtyp}\n\n"
            opts["tce__"] = "dft"
        else:
            mdccmd = f"task dft {runtyp}\n\n"

    else:
        raise InputError(f"Method not recognized: {method}")

    return mdccmd, opts



"""
Calls adcc
"""
from typing import TYPE_CHECKING, Dict

from qcelemental.models import AtomicResult, BasisSet, Provenance
from qcelemental.util import safe_version, which_import

from ..exceptions import InputError, UnknownError
from .model import ProgramHarness
from .qcvar_identities_resources import build_atomicproperties

if TYPE_CHECKING:
    from qcelemental.models import AtomicInput

    from ..config import TaskConfig

#


class AdccHarness(ProgramHarness):
    _defaults = {
        "name": "adcc",
        "scratch": False,
        "thread_safe": False,
        "thread_parallel": True,
        "node_parallel": False,
        "managed_memory": True,
    }
    version_cache: Dict[str, str] = {}

    class Config(ProgramHarness.Config):
        pass

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        """Whether adcc harness is ready for operation.
        Parameters
        ----------
        raise_error: bool
            Passed on to control negative return between False and ModuleNotFoundError raised.
        Returns
        -------
        bool
            If adcc and psi4 are found, returns True.
            If raise_error is False and adcc or psi4 is missing, returns False.
            If raise_error is True and adcc or psi4 are missing, the error message is raised.
        """
        found_adcc = which_import(
            "adcc",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install adcc -c adcc`.",
        )
        found_psi4 = which_import(
            "psi4",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install psi4 for adcc harness via `conda install psi4 -c psi4`.",
        )
        return found_adcc and found_psi4

    def get_version(self) -> str:
        """Return the currently used version of adcc"""
        self.found(raise_error=True)

        which_prog = which_import("adcc")
        if which_prog not in self.version_cache:
            import adcc

            self.version_cache[which_prog] = safe_version(adcc.__version__)
        return self.version_cache[which_prog]

    def compute(self, input_model: "AtomicInput", config: "TaskConfig") -> "AtomicResult":
        """
        Runs adcc
        """
        self.found(raise_error=True)

        import adcc
        import psi4

        mol = input_model.molecule
        model = input_model.model
        conv_tol = input_model.keywords.get("conv_tol", 1e-6)

        if input_model.driver not in ["energy", "properties"]:
            raise InputError(f"Driver {input_model.driver} not implemented for ADCC.")

        if isinstance(input_model.model.basis, BasisSet):
            raise InputError("QCSchema BasisSet for model.basis not implemented. Use string basis name.")
        if not input_model.model.basis:
            raise InputError("Model must contain a basis set.")

        psi4_molecule = psi4.core.Molecule.from_schema(dict(mol.dict(), fix_symmetry="c1"))
        psi4.core.clean()
        psi4.core.be_quiet()
        psi4.set_options(
            {
                "basis": model.basis,
                "scf_type": "pk",
                "e_convergence": conv_tol / 100,
                "d_convergence": conv_tol / 10,
                # 'maxiter': max_iter,
                "reference": "RHF" if mol.molecular_multiplicity == 1 else "UHF",
            }
        )
        _, wfn = psi4.energy("HF", return_wfn=True, molecule=psi4_molecule)
        adcc.set_n_threads(config.ncores)
        compute_success = False
        try:
            adcc_state = adcc.run_adc(wfn, method=model.method, **input_model.keywords)
            compute_success = adcc_state.converged
        except adcc.InputError as e:
            raise InputError(str(e))
        except Exception as e:
            raise UnknownError(str(e))

        input_data = input_model.dict(encoding="json")
        output_data = input_data.copy()
        output_data["success"] = compute_success

        if compute_success:
            output_data["return_result"] = adcc_state.excitation_energy[0]

            extract_props = input_model.driver == "properties"
            qcvars = adcc_state.to_qcvars(recurse=True, properties=extract_props)
            atprop = build_atomicproperties(qcvars)
            output_data["extras"]["qcvars"] = qcvars
            output_data["properties"] = atprop

        provenance = Provenance(creator="adcc", version=self.get_version(), routine="adcc").dict()
        provenance["nthreads"] = adcc.get_n_threads()
        output_data["provenance"] = provenance

        return AtomicResult(**output_data)



"""Compute geometrical counterpoise correction using Kruse & Grimme's GCP executable."""

import os
import pathlib
import pprint
import re
import socket
import sys
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

import numpy as np
import qcelemental as qcel
from qcelemental.models import AtomicResult, FailedOperation, Provenance
from qcelemental.util import safe_version, which

from ..exceptions import InputError, UnknownError
from ..util import execute
from .model import ProgramHarness

if TYPE_CHECKING:
    from qcelemental.models import AtomicInput

    from ..config import TaskConfig


pp = pprint.PrettyPrinter(width=120, compact=True, indent=1)


class GCPHarness(ProgramHarness):

    _defaults = {
        "name": "GCP",
        "scratch": True,
        "thread_safe": True,
        "thread_parallel": False,
        "node_parallel": False,
        "managed_memory": False,
    }
    version_cache: Dict[str, str] = {}

    class Config(ProgramHarness.Config):
        pass

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        return which(
            "gcp",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install gcp -c psi4`.",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        version = None
        which_prog = which("gcp")
        if which_prog not in self.version_cache:
            # option not (yet) available, instead find in help output
            command = [which_prog, "-version"]
            import subprocess

            proc = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            for ln in proc.stdout.decode("utf-8").splitlines():
                if re.match(".*Version", ln):
                    version = ln.split()[2]

            if version is None:
                raise UnknownError(f"Could not identify GCP version")
            # pending upcoming gcp update
            # self.version_cache[which_prog] = safe_version(
            #     proc.stdout.decode("utf-8").strip())
            self.version_cache[which_prog] = safe_version(version)

        return self.version_cache[which_prog]

    def compute(self, input_model: "AtomicInput", config: "TaskConfig") -> "AtomicResult":
        self.found(raise_error=True)

        job_inputs = self.build_input(input_model, config)

        success, dexe = self.execute(job_inputs)

        if success:
            dexe["outfiles"]["stdout"] = dexe["stdout"]
            dexe["outfiles"]["stderr"] = dexe["stderr"]
            dexe["outfiles"]["input"] = " ".join(job_inputs["command"])  # stretch to call this input
            dexe["outfiles"]["gcp_geometry.xyz"] = job_inputs["infiles"]["gcp_geometry.xyz"]
            output_model = self.parse_output(dexe["outfiles"], input_model)

        else:
            output_model = FailedOperation(
                success=False,
                error={"error_type": "execution_error", "error_message": dexe["stderr"]},
                input_data=input_model.dict(),
            )

        return output_model

    def execute(
        self, inputs: Dict[str, Any], *, extra_outfiles=None, extra_commands=None, scratch_name=None, timeout=None
    ) -> Tuple[bool, Dict]:

        success, dexe = execute(
            inputs["command"],
            inputs["infiles"],
            inputs["outfiles"],
            scratch_messy=inputs["scratch_messy"],
            scratch_directory=inputs["scratch_directory"],
            blocking_files=inputs["blocking_files"],
        )
        return success, dexe

    def build_input(
        self, input_model: "AtomicInput", config: "TaskConfig", template: Optional[str] = None
    ) -> Dict[str, Any]:

        if (input_model.driver.derivative_int() > 1) or (input_model.driver == "properties"):
            raise InputError(f"Driver {input_model.driver} not implemented for GCP.")

        # live somewhere else?
        available_levels = [
            "HF/MINIS",
            "DFT/MINIS",
            "HF/MINIX",
            "DFT/MINIX",
            "HF/SV",
            "DFT/SV",
            "HF/def2-SV(P)",
            "DFT/def2-SV(P)",
            "HF/def2-SVP",
            "DFT/def2-SVP",
            "HF/DZP",
            "DFT/DZP",
            "HF/def-TZVP",
            "DFT/def-TZVP",
            "HF/def2-TZVP",
            "DFT/def2-TZVP",
            "HF/631Gd",
            "DFT/631Gd",
            "HF/cc-pVDZ",
            "DFT/cc-pVDZ",
            "HF/aug-cc-pVDZ",
            "DFT/aug-cc-pVDZ",
            # DFT only
            "DFT/SVX",  # = def2-SV(P/h,c)
            "DFT/LANL",
            "DFT/pobTZVP",
            # HF only, needs update
            # 'HF/631G',
            # functional specific
            "TPSS/def2-SVP",
            "PW6B95/def2-SVP",
            # 3c specials
            "hf3c",
            "pbeh3c",
            # custom
            # will need gcp code changes if $HOME is to be avoided
            # thus code blocks with FILE below are not used yet.
            # 'file',
        ]
        available_levels = [f.upper() for f in available_levels]
        # temp until actual options object
        method = input_model.model.method.upper()
        if method not in available_levels:
            raise InputError(f"GCP does not have method: {method}")

        # Need 'real' field later and that's only guaranteed for molrec
        molrec = qcel.molparse.from_schema(input_model.molecule.dict())

        executable = self._defaults["name"].lower()
        calldash = {"gcp": "-", "mctc-gcp": "--"}[executable]

        command = [executable, "gcp_geometry.xyz", calldash + "level", method]

        if input_model.driver == "gradient":
            command.append(calldash + "grad")

        infiles = {
            "gcp_geometry.xyz": qcel.molparse.to_string(molrec, dtype="xyz", units="Angstrom", ghost_format=""),
        }
        if method == "FILE":
            infiles[".gcppar"] = input_model.extras["parameters"]

        return {
            "command": command,
            "infiles": infiles,
            "outfiles": ["gcp_gradient"],
            "scratch_messy": config.scratch_messy,
            "scratch_directory": config.scratch_directory,
            "input_result": input_model.copy(deep=True),
            "blocking_files": [os.path.join(pathlib.Path.home(), ".gcppar." + socket.gethostname())],
        }

    def parse_output(self, outfiles: Dict[str, str], input_model: "AtomicInput") -> "AtomicResult":
        stdout = outfiles.pop("stdout")
        stderr = outfiles.pop("stderr")

        # parse energy output (could go further and break into E6, E8, E10 and Cn coeff)
        real = np.array(input_model.molecule.real)
        full_nat = real.shape[0]
        real_nat = np.sum(real)

        for ln in stdout.splitlines():
            if re.match("  Egcp:", ln):
                ene = Decimal(ln.split()[1])
            elif re.match("     normal termination of gCP", ln):
                break
        else:
            if self._defaults["name"] == "GCP" and not ((real_nat == 1) and (input_model.driver == "gradient")):
                raise UnknownError(
                    f"Unsuccessful run. Check input, particularly geometry in [a0]. Model: {input_model.model}"
                )

        # parse gradient output
        if outfiles["gcp_gradient"] is not None:
            srealgrad = outfiles["gcp_gradient"].replace("D", "E")
            realgrad = np.fromstring(srealgrad, count=3 * real_nat, sep=" ").reshape((-1, 3))
        elif real_nat == 1:
            realgrad = np.zeros((1, 3))

        if input_model.driver == "gradient":
            ireal = np.argwhere(real).reshape((-1))
            fullgrad = np.zeros((full_nat, 3))
            try:
                fullgrad[ireal, :] = realgrad
            except NameError as exc:
                raise UnknownError("Unsuccessful gradient collection.") from exc

        qcvkey = input_model.model.method.upper()

        calcinfo = []

        calcinfo.append(qcel.Datum("CURRENT ENERGY", "Eh", ene))
        calcinfo.append(qcel.Datum("GCP CORRECTION ENERGY", "Eh", ene))
        if qcvkey:
            calcinfo.append(qcel.Datum(f"{qcvkey} GCP CORRECTION ENERGY", "Eh", ene))

        if input_model.driver == "gradient":
            calcinfo.append(qcel.Datum("CURRENT GRADIENT", "Eh/a0", fullgrad))
            calcinfo.append(qcel.Datum("GCP CORRECTION GRADIENT", "Eh/a0", fullgrad))
            if qcvkey:
                calcinfo.append(qcel.Datum(f"{qcvkey} GCP CORRECTION GRADIENT", "Eh/a0", fullgrad))

        calcinfo = {info.label: info.data for info in calcinfo}

        # Decimal --> str preserves precision
        calcinfo = {k.upper(): str(v) if isinstance(v, Decimal) else v for k, v in calcinfo.items()}

        retres = calcinfo[f"CURRENT {input_model.driver.upper()}"]
        if isinstance(retres, Decimal):
            retres = float(retres)
        elif isinstance(retres, np.ndarray):
            retres = retres.ravel().tolist()

        output_data = {
            "extras": input_model.extras,
            "native_files": {k: v for k, v in outfiles.items() if v is not None},
            "properties": {},
            "provenance": Provenance(
                creator="GCP", version=self.get_version(), routine=__name__ + "." + sys._getframe().f_code.co_name
            ),
            "return_result": retres,
            "stderr": stderr,
            "stdout": stdout,
        }

        output_data["extras"]["qcvars"] = calcinfo

        output_data["success"] = True
        return AtomicResult(**{**input_model.dict(), **output_data})


class MCTCGCPHarness(GCPHarness):

    _defaults = {
        "name": "MCTC-GCP",
        "scratch": True,
        "thread_safe": True,
        "thread_parallel": False,
        "node_parallel": False,
        "managed_memory": False,
    }
    version_cache: Dict[str, str] = {}

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        return which(
            "mctc-gcp",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install gcp-correction -c conda-forge`.",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which("mctc-gcp")
        if which_prog not in self.version_cache:
            command = [which_prog, "--version"]
            import subprocess

            proc = subprocess.run(command, stdout=subprocess.PIPE)
            self.version_cache[which_prog] = safe_version(proc.stdout.decode("utf-8").strip().split()[-1])

        return self.version_cache[which_prog]



"""
Tests the QCEngine compute module configuration
"""

import copy

import pydantic
import pytest

import qcengine as qcng
from qcengine.config import NodeDescriptor
from qcengine.util import create_mpi_invocation, environ_context


def test_node_blank():
    node = qcng.config.NodeDescriptor(name="something", hostname_pattern="*")


def test_node_auto():

    desc = {
        "name": "something",
        "hostname_pattern": "*",
        "jobs_per_node": 1,
        "ncores": 4,
        "memory": 10,
        "memory_safety_factor": 0,
    }
    node1 = qcng.config.NodeDescriptor(**desc)
    job1 = qcng.get_config(hostname=node1)
    assert job1.ncores == 4
    assert job1.nnodes == 1
    assert pytest.approx(job1.memory) == 10.0

    desc["jobs_per_node"] = 2
    node2 = qcng.config.NodeDescriptor(**desc)
    job2 = qcng.get_config(hostname=node2)
    assert job2.ncores == 2
    assert pytest.approx(job2.memory) == 5.0


def test_node_environ():

    scratch_name = "myscratch1234"
    with environ_context(env={"QCA_SCRATCH_DIR": scratch_name}):
        description = {"name": "something", "hostname_pattern": "*", "scratch_directory": "$QCA_SCRATCH_DIR"}

        node = qcng.config.NodeDescriptor(**description)
        assert node.scratch_directory == scratch_name


def test_node_skip_environ():
    description = {"name": "something", "hostname_pattern": "*", "scratch_directory": "$RANDOM_ENVIRON"}

    node = qcng.config.NodeDescriptor(**description)
    assert node.scratch_directory is None


@pytest.fixture
def opt_state_basic():
    """
    Capture the options state and temporarily override.
    """

    # Snapshot env
    old_node = copy.deepcopy(qcng.config.NODE_DESCRIPTORS)

    scratch_name = "myscratch1234"
    with environ_context(env={"QCA_SCRATCH_DIR": scratch_name}):

        configs = [
            {
                "name": "dragonstooth",
                "hostname_pattern": "dt*",
                "jobs_per_node": 2,
                "ncores": 12,
                "memory": 120,
                "scratch_directory": "$NOVAR_RANDOM_ABC123",
            },
            {"name": "newriver", "hostname_pattern": "nr*", "jobs_per_node": 2, "ncores": 24, "memory": 240},
            {
                "name": "batchnode",
                "hostname_pattern": "bn*",
                "is_batch_node": True,
                "jobs_per_node": 1,
                "mpiexec_command": "mpirun -n {total_ranks} -N {ranks_per_node}",
                "ncores": 24,
                "memory": 240,
            },
            {
                "name": "default",
                "hostname_pattern": "*",
                "jobs_per_node": 1,
                "memory": 4,
                "memory_safety_factor": 0,
                "ncores": 5,
                "scratch_directory": "$QCA_SCRATCH_DIR",
            },
        ]
        for desc in configs:
            node = qcng.config.NodeDescriptor(**desc)
            qcng.config.NODE_DESCRIPTORS[desc["name"]] = node

        yield

        # Reset env
        qcng.config.NODE_DESCRIPTORS = old_node


def test_node_matching(opt_state_basic):
    node = qcng.config.get_node_descriptor("nomatching")
    assert node.name == "default"

    node = qcng.config.get_node_descriptor("dt149")
    assert node.name == "dragonstooth"

    node = qcng.config.get_node_descriptor("nr149")
    assert node.name == "newriver"


def test_node_env(opt_state_basic):
    node = qcng.config.get_node_descriptor("dt")
    assert node.name == "dragonstooth"
    assert node.scratch_directory is None

    node = qcng.config.get_node_descriptor("nomatching")
    assert node.name == "default"
    assert node.scratch_directory == "myscratch1234"


def test_config_default(opt_state_basic):
    config = qcng.config.get_config(hostname="something")
    assert config.ncores == 5
    assert config.memory == 4

    config = qcng.config.get_config(hostname="dt149")
    assert config.ncores == 6
    assert config.retries == 0
    assert pytest.approx(config.memory, 0.1) == 54


def test_config_local_ncores(opt_state_basic):
    config = qcng.config.get_config(hostname="something", task_config={"ncores": 10, "retries": 3})
    assert config.ncores == 10
    assert config.memory == 4
    assert config.retries == 3


def test_config_local_njobs(opt_state_basic):
    config = qcng.config.get_config(hostname="something", task_config={"jobs_per_node": 5})
    assert config.ncores == 1
    assert pytest.approx(config.memory) == 0.8


def test_config_local_njob_ncore(opt_state_basic):
    config = qcng.config.get_config(hostname="something", task_config={"jobs_per_node": 3, "ncores": 1})
    assert config.ncores == 1
    assert pytest.approx(config.memory, 0.1) == 1.33


def test_config_local_njob_ncore_plus_memory(opt_state_basic):
    config = qcng.config.get_config(hostname="something", task_config={"jobs_per_node": 3, "ncores": 1, "memory": 6})
    assert config.ncores == 1
    assert pytest.approx(config.memory, 0.1) == 6


def test_config_local_nnodes(opt_state_basic):
    # Give a warning that mentions that mpirun is needed if you define a multi-node task
    with pytest.raises(ValueError) as exc:
        qcng.config.get_config(hostname="something", task_config={"nnodes": 10})
    assert "https://qcengine.readthedocs.io/en/stable/environment.html" in str(exc.value)

    # Test with an MPI run command
    local_options = {
        "nnodes": 4,
        "mpiexec_command": "mpirun -n {total_ranks} -N {ranks_per_node} --cpus-per-slot {cores_per_rank}",
    }
    config = qcng.config.get_config(hostname="something", task_config=local_options)
    assert config.use_mpiexec
    assert config.mpiexec_command.startswith("mpirun")
    assert create_mpi_invocation("hello_world", config) == [
        "mpirun",
        "-n",
        "20",
        "-N",
        "5",
        "--cpus-per-slot",
        "1",
        "hello_world",
    ]

    # Change the number of cores per rank
    local_options["cores_per_rank"] = 2
    config = qcng.config.get_config(hostname="something", task_config=local_options)
    assert config.use_mpiexec
    assert config.mpiexec_command.startswith("mpirun")
    assert create_mpi_invocation("hello_world", config) == [
        "mpirun",
        "-n",
        "10",
        "-N",
        "2",
        "--cpus-per-slot",
        "2",
        "hello_world",
    ]


def test_config_validation(opt_state_basic):
    with pytest.raises(pydantic.ValidationError):
        config = qcng.config.get_config(hostname="something", task_config={"bad": 10})


def test_global_repr():
    assert isinstance(qcng.config.global_repr(), str)


def test_batch_node(opt_state_basic):
    # Should always use mpirun
    config = qcng.config.get_config(hostname="bn1")
    assert config.use_mpiexec
    assert config.ncores == 24
    assert config.nnodes == 1

    config = qcng.config.get_config(hostname="bn1", task_config={"nnodes": 2})
    assert config.use_mpiexec
    assert config.ncores == 24
    assert config.nnodes == 2


def test_mpirun_command_errors():
    # Checks for ranks_per_node
    with pytest.raises(ValueError) as exc:
        NodeDescriptor(name="something", hostname_pattern="*", mpiexec_command="mpirun -n {total_ranks}")
    assert "must explicitly state" in str(exc.value)

    with pytest.raises(ValueError) as exc:
        NodeDescriptor(name="something", hostname_pattern="*", mpiexec_command="mpirun -N {ranks_per_node}")
    assert "must contain either" in str(exc.value)



from typing import Any, Dict

from qcengine.exceptions import InputError


def muster_modelchem(method: str, driver: "DriverEnum") -> Dict[str, Any]:
    """Converts the QC method into CFOUR keywords."""

    method = method.lower()
    opts = {}

    if driver == "properties":
        pass
    else:
        derint = driver.derivative_int()

        if derint == 0:
            if method == "cfour":
                pass  # permit clean operation of sandwich mode
            else:
                opts["deriv_level"] = "zero"

        elif derint == 1:
            opts["deriv_level"] = "first"

        elif derint == 2:
            opts["vibration"] = "exact"

    if method == "cfour":
        pass

    elif method in ["scf", "hf"]:
        opts["calc_level"] = "scf"

    elif method == "mp2":
        opts["calc_level"] = "mp2"

    elif method == "mp3":
        opts["calc_level"] = "mp3"

    elif method == "mp4(sdq)":
        opts["calc_level"] = "sdq-mp4"

    elif method == "mp4":
        opts["calc_level"] = "mp4"

    elif method == "bccd":
        opts["calc_level"] = "b-ccd"

    elif method == "cc2":
        opts["calc_level"] = "cc2"

    elif method == "ccsd":
        opts["calc_level"] = "ccsd"

    elif method == "cc3":
        opts["calc_level"] = "cc3"

    elif method == "ccsd(t)":
        # Can't use (T) b/c bug in xsymcor lops it off
        opts["calc_level"] = "ccsd[t]"

    elif method == "bccd(t)":
        opts["calc_level"] = "b-ccd(t)"

    elif method == "ccsdt":
        opts["calc_level"] = "ccsdt"

    else:
        raise InputError(f"Method not recognized: {method}")

    return opts



"""
Calls the Turbomole executable.
"""
import os
import re
from decimal import Decimal
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

from qcelemental.models import AtomicResult, BasisSet, Provenance
from qcelemental.util import safe_version, which

from ...exceptions import InputError
from ...util import execute, temporary_directory
from ..model import ProgramHarness
from ..qcvar_identities_resources import build_atomicproperties, build_out
from .define import execute_define, prepare_stdin
from .harvester import harvest
from .methods import KEYWORDS, METHODS


class TurbomoleHarness(ProgramHarness):

    _defaults = {
        "name": "Turbomole",
        "scratch": True,
        "thread_safe": False,
        "thread_parallel": False,
        "node_parallel": True,
        "managed_memory": True,
    }

    version_cache: Dict[str, str] = {}

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        return which(
            "define",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via http://www.cosmologic.de/turbomole/home.html",
        )

    def get_version(self) -> str:
        which_prog = which("define")
        if which_prog not in self.version_cache:
            # We use basically a dummy stdin as we dont want to pipe any real
            # input into define. We only want to parse the version number from
            # the string.
            with temporary_directory(suffix="_define_scratch") as tmpdir:
                tmpdir = Path(tmpdir)
                stdout = execute_define("\n", cwd=tmpdir)
            # Tested with V7.3 and V7.4.0
            version_re = re.compile("TURBOMOLE (?:rev\. )?(V.+?)\s+")
            mobj = version_re.search(stdout)
            version = mobj[1]
            self.version_cache[which_prog] = safe_version(version)
        return self.version_cache[which_prog]

    def compute(self, input_model: "AtomicInput", config: "TaskConfig") -> "AtomicResult":
        self.found(raise_error=True)

        job_inputs = self.build_input(input_model, config)
        success, dexe = self.execute(job_inputs)

        # TODO: handle input errors?! But then define probably already crashed...
        # if 'There is an error in the input file' in dexe["stdout"]:
        # raise InputError(dexe["stdout"])

        if success:
            dexe["outfiles"]["stdout"] = dexe["stdout"]
            dexe["outfiles"]["stderr"] = dexe["stderr"]
            return self.parse_output(dexe["outfiles"], input_model)

    def sub_control(self, control, pattern, repl, **kwargs):
        control_subbed = re.sub(pattern, repl, control, **kwargs)
        return control_subbed

    def append_control(self, control, to_append):
        return self.sub_control(control, "\$end", f"{to_append}\n$end")

    def build_input(
        self, input_model: "AtomicInput", config: "TaskConfig", template: Optional[str] = None
    ) -> Dict[str, Any]:

        # The 'define' wrapper can only handle normal string basis set input. If
        # a QCSchema basis set is given we break early, because this is not handled
        # right now.
        if isinstance(input_model.model.basis, BasisSet):
            raise InputError("QCSchema BasisSet for model.basis not implemented. Use string basis name.")

        turbomolerec = {
            "infiles": {},
            "outfiles": {"control": "control"},
            "scratch_directory": config.scratch_directory,
        }

        # Handle molecule
        # TODO: what's up with moldata? Do I need it?
        coord_str, moldata = input_model.molecule.to_string(dtype="turbomole", return_data=True)

        # Prepare stdin for define call
        model = input_model.model
        # geeopt will hold the for which to calculate the gradient.
        # 'x' corresponds to the ground state, 'a 1' would be the GS too.
        # 'a1 2' would be the 1st excited state of the irreducible group A1.
        # Right now only GS are supported, so this is hardcoded as 'x'.
        geoopt = "x" if input_model.driver.derivative_int() > 0 else ""
        stdin, subs = prepare_stdin(
            model.method,
            model.basis,
            input_model.keywords,
            input_model.molecule.molecular_charge,
            input_model.molecule.molecular_multiplicity,
            geoopt,
        )
        with temporary_directory(suffix="_define_scratch") as tmpdir:
            tmpdir = Path(tmpdir)
            with open(tmpdir / "coord", "w") as handle:
                handle.write(coord_str)
            stdout = execute_define(stdin, cwd=tmpdir)
            # The define scratch will be populated by some files that we want to keep
            to_keep = "basis auxbasis coord control alpha beta mos".split()

            for fn in to_keep:
                full_fn = tmpdir / fn
                if not full_fn.exists():
                    continue
                with open(full_fn) as handle:
                    turbomolerec["infiles"][fn] = handle.read()

        env = os.environ.copy()
        env["PARA_ARCH"] = "SMP"
        env["PARNODES"] = str(config.ncores)
        env["SMPCPUS"] = str(config.ncores)
        turbomolerec["environment"] = env
        # Memory is set in the control file

        keywords = input_model.keywords

        ########################
        # DETERMINE SOME FLAGS #
        ########################

        ri_calculation = any([keywords.get(ri_kw, False) for ri_kw in KEYWORDS["ri"]])
        ricc2_calculation = model.method in METHODS["ricc2"]

        ###################
        # MEMORY HANDLING #
        ###################

        # Central file that controls Turbomole. We assign it here to the "control"
        # variable as we may need to modify it, e.g. for a Hessian calculation.
        control = turbomolerec["infiles"]["control"]

        # Calculate total available memory in MB
        mem_mb = config.memory * (1024**3) / 1e6
        ri_fraction = 0.25
        # Total amount of memory allocated to ricore
        ricore = 0
        if ri_calculation:
            # This is the default given by Turbomole
            ricore = mem_mb * ri_fraction
            ri_per_core = int(ricore / config.ncores)
            # Update $ricore entry in the control file
            control = self.sub_control(control, "\$ricore\s+(\d+)", f"$ricore {ri_per_core} MiB per_core")
        # Calculate remaining memory
        maxcor = mem_mb - ricore
        assert maxcor > 0, "Not enough memory for maxcor! Need {-maxcor} MB more!"

        # maxcore per_core
        per_core = int(maxcor / config.ncores)
        # Update $maxcor entry in the control file
        control = self.sub_control(control, "\$maxcor\s+(\d+)\s+MiB\s+per_core", f"$maxcor {per_core} MiB per_core")

        ############################
        # DETERMINE SHELL COMMANDS #
        ############################

        # ----------------------#
        # | Energy calculations |
        # ----------------------#

        # Set appropriate commands. We always need a reference wavefunction
        # so the first command will be dscf or ridft to converge the SCF.
        commands = ["ridft"] if ri_calculation else ["dscf"]

        # ------------------------#
        # | Gradient calculations |
        # ------------------------#

        # Keep the gradient file for parsing
        if input_model.driver.derivative_int() == 1:
            turbomolerec["outfiles"]["gradient"] = "gradient"

        # ricc2 will also calculate the gradient. But this requires setting
        # 'geoopt (state)' in the control file. This is currently handled in the
        # 'define' call.
        if ricc2_calculation:
            commands.append("ricc2")
        # Gradient calculation for DFT/HF
        elif input_model.driver.derivative_int() == 1:
            grad_command = "rdgrad" if ri_calculation else "grad"
            commands.append(grad_command)

        # -----------------------#
        # | Hessian calculations |
        # -----------------------#

        if input_model.driver.derivative_int() == 2:
            freq_command = "NumForce -level cc2" if ricc2_calculation else "aoforce"
            # NumForce seems to ignore the nprhessian command and will always
            # write to unprojected and projected Hessian to "hessian".
            hessian_outfile = "hessian" if ricc2_calculation else "nprhessian"
            commands.append(freq_command)
            # Add some keywords to the control file
            #   noproj: Don't project out translation and rotation
            #   nprhessian: Set filename of un-projected hessian
            control = self.append_control(control, "$noproj\n$nprhessian file=nprhessian")
            turbomolerec["outfiles"][hessian_outfile] = None

        # Build the full shell command and set it
        command = ["; ".join(commands)]
        turbomolerec["command"] = command
        # Re-assign the potentially modified control file, e.g. for a Hessian calculation
        turbomolerec["infiles"]["control"] = control

        # TODO: check if the chosen commands are available with which()?

        return turbomolerec

    def execute(
        self, inputs: Dict[str, Any], *, extra_outfiles=None, extra_commands=None, scratch_name=None, timeout=None
    ) -> Tuple[bool, Dict]:

        success, dexe = execute(
            inputs["command"],
            inputs["infiles"],
            inputs["outfiles"],
            shell=True,
            # TODO: scratch_messy?
            # scratch_messy=False,
        )
        return success, dexe

    def parse_output(
        self, outfiles: Dict[str, str], input_model: "AtomicInput"
    ) -> "AtomicResult":  # lgtm: [py/similar-function]

        stdout = outfiles.pop("stdout")

        qcvars, gradient, hessian = harvest(input_model.molecule, stdout, **outfiles)

        if gradient is not None:
            qcvars["CURRENT GRADIENT"] = gradient

        if hessian is not None:
            qcvars["CURRENT HESSIAN"] = hessian

        retres = qcvars[f"CURRENT {input_model.driver.upper()}"]
        if isinstance(retres, Decimal):
            retres = float(retres)

        build_out(qcvars)
        atprop = build_atomicproperties(qcvars)

        output_data = input_model.dict()
        output_data["extras"]["outfiles"] = outfiles
        output_data["properties"] = atprop
        output_data["provenance"] = Provenance(creator="Turbomole", version=self.get_version(), routine="turbomole")
        output_data["return_result"] = retres
        output_data["stdout"] = stdout
        output_data["success"] = True

        return AtomicResult(**output_data)



import numpy as np
import pytest
import qcelemental as qcel

from qcengine.programs.tests.standard_suite_ref import std_molecules, std_refs
from qcengine.testing import using

from .standard_suite_runner import runner_asserter
from .test_standard_suite import _processor


@pytest.fixture
def clsd_open_pmols():
    frame_not_important = {
        name[:-4]: qcel.models.Molecule.from_data(smol, name=name[:-4])
        for name, smol in std_molecules.items()
        if name.endswith("-xyz")
    }
    frame_part_of_spec = {
        name[:-4] + "-fixed": qcel.models.Molecule.from_data(smol + "\nno_com\nno_reorient\n", name=name[:-4])
        for name, smol in std_molecules.items()
        if name.endswith("-xyz")
    }

    return {**frame_not_important, **frame_part_of_spec}


#
#  ,--.  ,--.,------.      ,---.  ,--.,--.                                          ,--.
#  |  '--'  ||  .---'     /  O  \ |  |`--' ,---. ,--,--, ,--,--,--. ,---. ,--,--, ,-'  '-.
#  |  .--.  ||  `--,     |  .-.  ||  |,--.| .-. ||      \|        || .-. :|      \'-.  .-'
#  |  |  |  ||  |`       |  | |  ||  ||  |' '-' '|  ||  ||  |  |  |\   --.|  ||  |  |  |
#  `--'  `--'`--'        `--' `--'`--'`--'.`-  / `--''--'`--`--`--' `----'`--''--'  `--'
#                                         `---'
#  <<<  HF Alignment


@pytest.mark.parametrize(
    "scramble",
    # * this parameter alters the input molecule by Cartesian frame and/or atom ordering to imitate real-world inputs.
    # * scramble dictionary are arguments to qcdb.Molecule.scramble() or qcel.models.Molecule.scramble() that computes a shifted/rotated/atom-mapped input molecule from `subjects` below and the transformations to be applied to the reference data.
    # * arguments do_shift=T/F and other boolean values generate random perturbations, so multiple lines or runs makes a fuller test.
    # * specific shifts etc or problematic runs can be reproduced by specifying arrays like the commented example below. note that b/c do_resort is natom-dependent, may need to exclude subjects while debugging.
    # * `id`s below are numbered since `pytest -k` doesn't recognize case and so duplicate entries can be added. the 0, 1, 2, 3, 9 progression is handy for debugging since perturbations are added systematically
    [
        # pytest.param({"do_shift": [ 2.660760432055,  1.477336796939, -2.098045335573], "do_rotate": [[ 0.321861140022,  0.445246880671,  0.835560064745], [-0.447874157747,  0.849136107328, -0.279958229125], [-0.834154749052, -0.28411808546 ,  0.472718487209]], "do_resort": [2, 0, 1]}, id="mTTT3"),
        # pytest.param({"do_shift": False, "do_rotate": False, "do_resort": False}, id="srm0"),
        # pytest.param({"do_shift": True, "do_rotate": False, "do_resort": False}, id="Srm1"),
        # pytest.param({"do_shift": True, "do_rotate": True, "do_resort": False}, id="SRm2"),
        # pytest.param({"do_shift": False, "do_rotate": False, "do_resort": True}, id="srM3"),
        # pytest.param({"do_shift": True, "do_rotate": True, "do_resort": True}, id="SRM8"),
        pytest.param({"do_shift": True, "do_rotate": True, "do_resort": True}, id="SRM9"),
    ],
)
@pytest.mark.parametrize(
    "frame",
    # * this parameter alters the input molecule by fix_com and fix_orientation to imitate user signaling frame matters or not.
    [
        pytest.param("fixed"),  # fix_=True (no_com/no_reorient); atres.mol.geom = atin.mol.geom aka scrambled
        pytest.param("free"),  # fix_=False (def)               ; atres.mol.geom = dsl internal orientation
    ],
)
@pytest.mark.parametrize(
    "driver",
    [
        pytest.param("energy", id="ene0"),
        pytest.param("gradient", id="grd1"),
        pytest.param("hessian", id="hes2"),
    ],
)
@pytest.mark.parametrize(
    "basis, subjects",
    # this parameter, along with rhf/uhf below covers four different molecules, with nat=2-4.
    [
        pytest.param("cc-pvdz", ["hf", "bh3p", "bh3p"], id="dz"),
        pytest.param("aug-cc-pvdz", ["h2o", "nh2", "nh2"], id="adz", marks=pytest.mark.long),
    ],
)
@pytest.mark.parametrize(
    "inp",
    [
        # yapf: disable
        pytest.param({"call": "cfour",  "reference": "rhf",  "fcae": "ae", "keywords": {"scf_conv": 12},                                                                     }, id="hf  rhf ae: cfour",      marks=using("cfour")),
        pytest.param({"call": "gamess", "reference": "rhf",  "fcae": "ae", "keywords": {},                                                                                   }, id="hf  rhf ae: gamess",     marks=using("gamess")),
        pytest.param({"call": "nwchem", "reference": "rhf",  "fcae": "ae", "keywords": {"scf__thresh": 1.e-6},                                                               }, id="hf  rhf ae: nwchem",     marks=using("nwchem")),
        pytest.param({"call": "psi4",   "reference": "rhf",  "fcae": "ae", "keywords": {"scf_type": "pk"},                                                                   }, id="hf  rhf ae: psi4",       marks=using("psi4_derqcsk")),

        pytest.param({"call": "cfour",  "reference": "uhf",  "fcae": "ae", "keywords": {"reference": "uhf", "scf_conv": 12},                                                 }, id="hf  uhf ae: cfour",      marks=using("cfour")),
        pytest.param({"call": "gamess", "reference": "uhf",  "fcae": "ae", "keywords": {"contrl__scftyp": "uhf"},                                                            }, id="hf  uhf ae: gamess",     marks=using("gamess")),
        pytest.param({"call": "nwchem", "reference": "uhf",  "fcae": "ae", "keywords": {"scf__uhf": True, "scf__thresh": 1.e-6},                                             }, id="hf  uhf ae: nwchem",     marks=using("nwchem")),
        pytest.param({"call": "psi4",   "reference": "uhf",  "fcae": "ae", "keywords": {"reference": "uhf", "scf_type": "pk"},                                               }, id="hf  uhf ae: psi4",       marks=using("psi4_derqcsk")),
        # yapf: enable
    ],
)
def test_hf_alignment(inp, scramble, frame, driver, basis, subjects, clsd_open_pmols, request):
    runner_asserter(
        *_processor(inp, "", basis, subjects, clsd_open_pmols, request, driver, "hf", scramble=scramble, frame=frame)
    )



import math
import re

import numpy as np
from qcelemental.exceptions import ValidationError
from qcelemental.util import filter_comments


def load_hessian(shess: str, dtype: str) -> np.ndarray:
    """Construct a Hessian array from any recognized string format.

    Parameters
    ----------
    shess
        Multiline string specification of Hessian in a recognized format.
    dtype
        {"fcmfinal", "cfour", "gamess"}
        Hessian format name.

    Returns
    -------
    np.ndarray
        Hessian array in square shape.

    """
    # list o'lines w/o comments or blanks
    shess = filter_comments(shess)
    lhess = list(filter(None, map(str.strip, shess.splitlines())))

    if dtype in ["fcmfinal", "cfour"]:
        nat = int(lhess[0].split()[0])
        ndof = 3 * nat
        datastr = "\n".join(lhess[1:])
        nhess = np.fromstring(datastr, sep=" ")
        nhess = nhess.reshape(ndof, ndof)
    elif dtype == "gamess":
        if "ENERGY" in lhess[0]:
            lhess.pop(0)
        datastr = []
        for ln in lhess:
            numbers = re.findall(r"([-+]?\d+\.\d+[DdEe][-+]\d\d)", ln)
            if numbers:
                datastr.extend(numbers)

        nhess = np.fromstring(" ".join(datastr), sep=" ")
        ndof = int(math.sqrt(len(nhess)))
        nhess = nhess.reshape((ndof, ndof))
    else:
        raise ValidationError("Unknown dtype: {}".format(dtype))

    return nhess


def hess_to_string(hess, handle, dtype):
    nat = hess.shape[0] // 3
    assert hess.shape == (3 * nat, 3 * nat)

    header = "{:5}{:5}".format(nat, 6 * nat)
    np.savetxt(handle, hess.reshape((-1, 3)), fmt="%20.10f", delimiter="", newline="\n", header=header, comments="")



import argparse
import os
import shutil
import subprocess as sp

# Args
parser = argparse.ArgumentParser(description='Creates a conda environment from file for a given Python version.')
parser.add_argument('-n', '--name', type=str, nargs=1, help='The name of the created Python environment')
parser.add_argument('-p', '--python', type=str, nargs=1, help='The version of the created Python environment')
parser.add_argument('conda_file', nargs='*', help='The file for the created Python environment')

args = parser.parse_args()

with open(args.conda_file[0], "r") as handle:
    script = handle.read()

tmp_file = "tmp_env.yaml"
script = script.replace("- python", "- python {}*".format(args.python[0]))

with open(tmp_file, "w") as handle:
    handle.write(script)

conda_path = shutil.which("conda")

print("CONDA ENV NAME  {}".format(args.name[0]))
print("PYTHON VERSION  {}".format(args.python[0]))
print("CONDA FILE NAME {}".format(args.conda_file[0]))
print("CONDA path      {}".format(conda_path))

sp.call("{} env create -n {} -f {}".format(conda_path, args.name[0], tmp_file), shell=True)
os.unlink(tmp_file)



"""
Base file for the dqm_compute module.
"""

from . import config, exceptions
from .compute import compute, compute_procedure
from .config import get_config
from .extras import get_information
from .mdi_server import MDIServer
from .procedures import get_procedure, list_all_procedures, list_available_procedures
from .programs import get_program, list_all_programs, list_available_programs, register_program, unregister_program
from .stock_mols import get_molecule

# Handle versioneer
__version__ = get_information("version")
__git_revision__ = get_information("git_revision")
del get_information



"""
Calls the Psi4 executable.
"""
import json
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Dict

from qcelemental.models import AtomicResult, BasisSet
from qcelemental.util import deserialize, parse_version, safe_version, which, which_import

from ..exceptions import InputError, RandomError, ResourceError, UnknownError
from ..util import execute, popen, temporary_directory
from .model import ProgramHarness

if TYPE_CHECKING:
    from qcelemental.models import AtomicInput

    from ..config import TaskConfig


class Psi4Harness(ProgramHarness):

    _defaults = {
        "name": "Psi4",
        "scratch": True,
        "thread_safe": False,
        "thread_parallel": True,
        "node_parallel": False,
        "managed_memory": True,
    }
    version_cache: Dict[str, str] = {}

    class Config(ProgramHarness.Config):
        pass

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        """Whether Psi4 harness is ready for operation.

        Parameters
        ----------
        raise_error: bool
            Passed on to control negative return between False and ModuleNotFoundError raised.

        Returns
        -------
        bool
            If psi4 (psithon or psiapi) is found, returns True.
            If raise_error is False and psi4 is missing, returns False.
            If raise_error is True and psi4 is missing, the error message is raised.

        """
        psithon = which("psi4", return_bool=True)
        psiapi = which_import("psi4", return_bool=True)

        if psithon and not psiapi:
            with popen([which("psi4"), "--module"]) as exc:
                exc["proc"].wait(timeout=30)
            if "module does not exist" in exc["stderr"]:
                pass  # --module argument only in Psi4 DDD branch
            else:
                sys.path.append(exc["stdout"].split()[-1])

        if psiapi and not psithon:
            psiimport = str(Path(which_import("psi4")).parent.parent)
            env = os.environ.copy()
            env["PYTHONPATH"] = psiimport
            with popen(["python", "-c", "import psi4; print(psi4.executable[:-5])"], popen_kwargs={"env": env}) as exc:
                exc["proc"].wait(timeout=30)
            os.environ["PATH"] += os.pathsep + exc["stdout"].split()[-1]

        if psithon or psiapi:
            return True

        return which(
            "psi4",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install psi4 -c psi4`. Check it's in your PATH with `which psi4`.",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which("psi4")
        if which_prog not in self.version_cache:
            with popen([which_prog, "--version"]) as exc:
                exc["proc"].wait(timeout=30)
            self.version_cache[which_prog] = safe_version(exc["stdout"].split()[-1])

        candidate_version = self.version_cache[which_prog]

        if "undef" in candidate_version:
            raise TypeError(
                """Using custom build without tags. Please pull git tags with `git pull origin master --tags`. If building from source, `git fetch upstream "refs/tags/*:refs/tags/*"`."""
            )

        return candidate_version

    def _handle_errors(self, output_data):
        if "error" in output_data:
            if "error_message" not in output_data["error"]:
                error_message = output_data["error"]
                error_type = "internal_error"
            else:
                error_message = output_data["error"]["error_message"]
                error_type = output_data["error"].get("error_type", "unknown_error")
        else:
            error_message = "Unknown error, error message is not found, possibly segfaulted"
            error_type = "internal_error"

        return error_message, error_type

    def compute(self, input_model: "AtomicInput", config: "TaskConfig") -> "AtomicResult":
        """
        Runs Psi4 in API mode
        """
        self.found(raise_error=True)
        pversion = parse_version(self.get_version())

        if pversion < parse_version("1.2"):
            raise ResourceError("Psi4 version '{}' not understood.".format(self.get_version()))

        # Location resolution order config.scratch_dir, $PSI_SCRATCH, /tmp
        parent = config.scratch_directory
        if parent is None:
            parent = os.environ.get("PSI_SCRATCH", None)

        error_type = None
        error_message = None
        compute_success = False

        if isinstance(input_model.model.basis, BasisSet):
            raise InputError("QCSchema BasisSet for model.basis not implemented. Use string basis name.")

        # Basis must not be None for HF3c
        old_basis = input_model.model.basis
        input_model.model.__dict__["basis"] = old_basis or ""

        with temporary_directory(parent=parent, suffix="_psi_scratch", messy=config.scratch_messy) as tmpdir:

            caseless_keywords = {k.lower(): v for k, v in input_model.keywords.items()}
            if (input_model.molecule.molecular_multiplicity != 1) and ("reference" not in caseless_keywords):
                input_model.keywords["reference"] = "uhf"

            # Old-style JSON-based command line
            if pversion < parse_version("1.4a2.dev160"):

                # Setup the job
                input_data = input_model.dict(encoding="json")
                input_data["nthreads"] = config.ncores
                input_data["memory"] = int(config.memory * 1024 * 1024 * 1024 * 0.95)  # Memory in bytes
                input_data["success"] = False
                input_data["return_output"] = True

                if input_data["schema_name"] == "qcschema_input":
                    input_data["schema_name"] = "qc_schema_input"

                # Execute the program
                success, output = execute(
                    [which("psi4"), "--scratch", tmpdir, "--json", "data.json"],
                    {"data.json": json.dumps(input_data)},
                    ["data.json"],
                    scratch_directory=tmpdir,
                )

                output_data = input_data.copy()
                if success:
                    output_data = json.loads(output["outfiles"]["data.json"])
                    if "extras" not in output_data:
                        output_data["extras"] = {}

                    # Check QCVars
                    local_qcvars = output_data.pop("psi4:qcvars", None)
                    if local_qcvars:
                        # Edge case where we might already have qcvars, should not happen
                        if "qcvars" in output_data["extras"]:
                            output_data["extras"]["local_qcvars"] = local_qcvars
                        else:
                            output_data["extras"]["qcvars"] = local_qcvars

                    if output_data.get("success", False) is False:
                        error_message, error_type = self._handle_errors(output_data)
                    else:
                        compute_success = True

                else:
                    error_message = output.get("stderr", "No STDERR output")
                    error_type = "execution_error"

                # Reset the schema if required
                output_data["schema_name"] = "qcschema_output"
                output_data.pop("memory", None)
                output_data.pop("nthreads", None)
                output_data["stdout"] = output_data.pop("raw_output", None)

            else:

                if input_model.extras.get("psiapi", False):
                    import psi4

                    orig_scr = psi4.core.IOManager.shared_object().get_default_path()
                    psi4.core.set_num_threads(config.ncores, quiet=True)
                    psi4.set_memory(f"{config.memory}GiB", quiet=True)
                    # psi4.core.IOManager.shared_object().set_default_path(str(tmpdir))
                    if pversion < parse_version("1.6a2"):  # adjust to where DDD merged
                        # slightly dangerous in that if `qcng.compute({..., psiapi=True}, "psi4")` called *from psi4
                        #   session*, session could unexpectedly get its own files cleaned away.
                        output_data = psi4.schema_wrapper.run_qcschema(input_model).dict()
                    else:
                        output_data = psi4.schema_wrapper.run_qcschema(input_model, postclean=False).dict()
                    # success here means execution returned. output_data may yet be qcel.models.AtomicResult or qcel.models.FailedOperation
                    success = True
                    if output_data.get("success", False):
                        output_data["extras"]["psiapi_evaluated"] = True
                    psi4.core.IOManager.shared_object().set_default_path(orig_scr)
                else:
                    run_cmd = [
                        which("psi4"),
                        "--scratch",
                        str(tmpdir),
                        "--nthread",
                        str(config.ncores),
                        "--memory",
                        f"{config.memory}GiB",
                        "--qcschema",
                        "data.msgpack",
                    ]
                    if config.scratch_messy:
                        run_cmd.append("--messy")
                    input_files = {"data.msgpack": input_model.serialize("msgpack-ext")}
                    success, output = execute(
                        run_cmd, input_files, ["data.msgpack"], as_binary=["data.msgpack"], scratch_directory=tmpdir
                    )
                    if success:
                        output_data = deserialize(output["outfiles"]["data.msgpack"], "msgpack-ext")
                    else:
                        output_data = input_model.dict()

                if success:
                    if output_data.get("success", False) is False:
                        error_message, error_type = self._handle_errors(output_data)
                    else:
                        compute_success = True
                else:
                    error_message = output.get("stderr", "No STDERR output")
                    error_type = "execution_error"

        # Dispatch errors, PSIO Errors are not recoverable for future runs
        if compute_success is False:

            if "PSIO Error" in error_message:
                if "scratch directory" in error_message:
                    # Psi4 cannot access the folder or file
                    raise ResourceError(error_message)
                else:
                    # Likely a random error, worth retrying
                    raise RandomError(error_message)
            elif ("SIGSEV" in error_message) or ("SIGSEGV" in error_message) or ("segmentation fault" in error_message):
                raise RandomError(error_message)
            elif (
                # Missing and Managed cover same category of error pre- and post-DDD
                ("TypeError: set_global_option" in error_message)
                or (error_type in ["ValidationError", "MissingMethodError", "ManagedMethodError"])
            ):
                raise InputError(error_message)
            elif "RHF reference is only for singlets" in error_message:
                raise InputError(error_message)
            else:
                raise UnknownError(error_message)

        # Reset basis
        output_data["model"]["basis"] = old_basis

        # Move several pieces up a level
        output_data["provenance"]["memory"] = round(config.memory, 3)
        output_data["provenance"]["nthreads"] = config.ncores
        if output_data.get("native_files", None) is None:
            output_data["native_files"] = {
                "input": json.dumps(json.loads(input_model.json()), indent=1),
            }

        # Delete keys
        output_data.pop("return_output", None)

        return AtomicResult(**output_data)
